from __future__ import annotations

import copy

import pytest

from daft.datatype import DataType

daft_int_types = [
    DataType.int8(),
    DataType.int16(),
    DataType.int32(),
    DataType.int64(),
    DataType.uint8(),
    DataType.uint16(),
    DataType.uint32(),
    DataType.uint64(),
]

daft_numeric_types = daft_int_types + [DataType.float32(), DataType.float64()]
daft_string_types = [DataType.string()]
daft_nonnull_types = (
    daft_numeric_types
    + daft_string_types
    + [DataType.bool(), DataType.binary(), DataType.fixed_size_binary(10), DataType.date()]
)

daft_list_types = [DataType.list(t) for t in daft_nonnull_types] + [
    DataType.fixed_size_list(t, size=10) for t in daft_nonnull_types
]
daft_embedding_types = [DataType.embedding(t, 512) for t in daft_numeric_types]
daft_tensor_types = [DataType.tensor(t) for t in daft_numeric_types]
daft_sparse_tensor_types = [DataType.sparse_tensor(t) for t in daft_numeric_types]
daft_fixed_shape_tensor_types = [DataType.tensor(t, shape=(2, 3)) for t in daft_numeric_types]
daft_fixed_shape_sparse_tensor_types = [DataType.sparse_tensor(t, shape=(2, 3)) for t in daft_numeric_types]
daft_map_types = [DataType.map(DataType.string(), t) for t in daft_nonnull_types]
daft_struct_types = [DataType.struct({"foo": t, "bar": t}) for t in daft_nonnull_types]

all_daft_types = (
    daft_nonnull_types
    + daft_list_types
    + daft_map_types
    + daft_struct_types
    + daft_embedding_types
    + daft_tensor_types
    + daft_sparse_tensor_types
    + daft_fixed_shape_tensor_types
    + daft_fixed_shape_sparse_tensor_types
)


@pytest.mark.parametrize("dtype", daft_nonnull_types)
def test_datatype_pickling(dtype) -> None:
    copy_dtype = copy.deepcopy(dtype)
    assert copy_dtype == dtype


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        (str, DataType.string()),
        (int, DataType.int64()),
        (float, DataType.float64()),
        (bytes, DataType.binary()),
        (object, DataType.python()),
        (
            {"foo": str, "bar": int},
            DataType.struct({"foo": DataType.string(), "bar": DataType.int64()}),
        ),
    ],
)
def test_datatype_parsing(source, expected):
    assert DataType._infer_type(source) == expected


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        # These tests must be run in later version of Python that allow for subscripting of types
        (list[str], DataType.list(DataType.string())),
        (dict[str, int], DataType.map(DataType.string(), DataType.int64())),
        (
            {"foo": list[str], "bar": int},
            DataType.struct({"foo": DataType.list(DataType.string()), "bar": DataType.int64()}),
        ),
        (list[list[str]], DataType.list(DataType.list(DataType.string()))),
    ],
)
def test_subscripted_datatype_parsing(source, expected):
    assert DataType._infer_type(source) == expected


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_null(test_type):
    assert test_type.is_null() == (test_type == DataType.null())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_boolean(test_type):
    assert test_type.is_boolean() == (test_type == DataType.bool())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_int8(test_type):
    assert test_type.is_int8() == (test_type == DataType.int8())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_int16(test_type):
    assert test_type.is_int16() == (test_type == DataType.int16())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_int32(test_type):
    assert test_type.is_int32() == (test_type == DataType.int32())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_int64(test_type):
    assert test_type.is_int64() == (test_type == DataType.int64())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_uint8(test_type):
    assert test_type.is_uint8() == (test_type == DataType.uint8())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_uint16(test_type):
    assert test_type.is_uint16() == (test_type == DataType.uint16())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_uint32(test_type):
    assert test_type.is_uint32() == (test_type == DataType.uint32())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_uint64(test_type):
    assert test_type.is_uint64() == (test_type == DataType.uint64())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_float32(test_type):
    assert test_type.is_float32() == (test_type == DataType.float32())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_float64(test_type):
    assert test_type.is_float64() == (test_type == DataType.float64())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_decimal128(test_type):
    assert test_type.is_decimal128() == (test_type == DataType.decimal128(10, 2))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_timestamp(test_type):
    assert test_type.is_timestamp() == (test_type == DataType.timestamp("ns"))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_date(test_type):
    assert test_type.is_date() == (test_type == DataType.date())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_time(test_type):
    assert test_type.is_time() == (test_type == DataType.time("ns"))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_duration(test_type):
    assert test_type.is_duration() == (test_type == DataType.duration("ns"))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_interval(test_type):
    assert test_type.is_interval() == (test_type == DataType.interval())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_binary(test_type):
    assert test_type.is_binary() == (test_type == DataType.binary())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_fixed_size_binary(test_type):
    assert test_type.is_fixed_size_binary() == (test_type == DataType.fixed_size_binary(10))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_string(test_type):
    assert test_type.is_string() == (test_type == DataType.string())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_list(test_type):
    if test_type.is_list():
        assert test_type.is_list() == (test_type == DataType.list(test_type.dtype))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_fixed_size_list(test_type):
    if test_type.is_fixed_size_list():
        assert test_type.is_fixed_size_list() == (test_type == DataType.fixed_size_list(test_type.dtype, 10))


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_image(test_type):
    assert test_type.is_image() == (test_type == DataType.image())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_fixed_shape_image(test_type):
    if test_type.is_fixed_shape_image():
        assert test_type == DataType.image(mode="RGB", height=224, width=224)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_embedding(test_type):
    if test_type.is_embedding():
        assert test_type == DataType.embedding(test_type.dtype, test_type.size)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_tensor(test_type):
    if test_type.is_tensor():
        assert test_type == DataType.tensor(test_type.dtype)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_fixed_shape_tensor(test_type):
    if test_type.is_fixed_shape_tensor():
        assert test_type == DataType.tensor(test_type.dtype, test_type.shape)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_sparse_tensor(test_type):
    if test_type.is_sparse_tensor():
        assert test_type == DataType.sparse_tensor(test_type.dtype)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_fixed_shape_sparse_tensor(test_type: DataType):
    if test_type.is_fixed_shape_sparse_tensor():
        test_type == DataType.sparse_tensor(test_type.dtype, shape=test_type.shape)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_python(test_type):
    assert test_type.is_python() == (test_type == DataType.python())


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_numeric(test_type):
    assert test_type.is_numeric() == (test_type in daft_numeric_types)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_integer(test_type):
    assert test_type.is_integer() == (test_type in daft_int_types)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_is_temporal(test_type):
    temporal_types = [
        DataType.timestamp("ns"),
        DataType.date(),
        DataType.time("ns"),
        DataType.duration("ns"),
        DataType.interval(),
    ]
    assert test_type.is_temporal() == (test_type in temporal_types)


@pytest.mark.parametrize("test_type", all_daft_types)
def test_fixed_size_property(test_type):
    if test_type.is_fixed_size_list() or test_type.is_fixed_size_binary():
        assert test_type.size == 10
    elif test_type.is_embedding():
        assert test_type.size == 512
    else:
        try:
            test_type.size
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_inner_type_property(test_type: DataType):
    if (
        test_type.is_list()
        or test_type.is_fixed_size_list()
        or test_type.is_tensor()
        or test_type.is_sparse_tensor()
        or test_type.is_fixed_shape_tensor()
        or test_type.is_fixed_shape_sparse_tensor()
        or test_type.is_embedding()
    ):
        try:
            assert test_type.dtype is not None
        except AttributeError:
            assert False


@pytest.mark.parametrize("test_type", all_daft_types)
def test_struct_fields_property(test_type: DataType):
    if test_type.is_struct():
        assert test_type.fields is not None
    else:
        try:
            test_type.fields
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_precision_and_scale_properties(test_type):
    if test_type.is_decimal128():
        assert test_type.precision == 10
        assert test_type.scale == 2
    else:
        try:
            test_type.precision
            test_type.scale
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_time_unit_property(test_type: DataType):
    if test_type.is_time() or test_type.is_duration() or test_type.is_timestamp():
        assert test_type.timeunit == "ns"
    else:
        try:
            test_type.timeunit
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_image_properties(test_type: DataType):
    if test_type.is_image() or test_type.is_fixed_shape_image():
        assert test_type.image_mode == "RGB"
    else:
        try:
            test_type.image_mode
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_use_offset_indices_property(test_type: DataType):
    if test_type.is_sparse_tensor() or test_type.is_fixed_shape_sparse_tensor():
        assert test_type.use_offset_indices is False
    else:
        try:
            test_type.use_offset_indices
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_map_key_property(test_type: DataType):
    if test_type.is_map():
        assert test_type.key_type == DataType.string()
    else:
        try:
            test_type.key_type
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True


@pytest.mark.parametrize("test_type", all_daft_types)
def test_map_value_property(test_type: DataType):
    if test_type.is_map():
        assert test_type.value_type is not None
    else:
        try:
            test_type.value_type
            pytest.fail("Expected AttributeError")
        except AttributeError:
            assert True



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    ["expr", "data"],
    [
        (col("col").str.find("oo"), ["foo", "quux"]),
        (
            col("col").str.find(lit("oo")),
            ["foo", "quux"],
        ),
        (
            col("col").str.find(col("emptystrings") + lit("oo")),
            ["foo", "quux"],
        ),
    ],
)
def test_series_utf8_find_broadcast_pattern(expr, data) -> None:
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", ""]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": [1, -1]}



from __future__ import annotations

import tempfile

import pyarrow as pa
import pyarrow.parquet as papq


def generate_parquet_file():
    """Generate a small Parquet file and return the path to the file."""
    tmpfile = tempfile.NamedTemporaryFile()
    tbl = pa.Table.from_pydict({"foo": [1, 2, 3]})
    papq.write_table(tbl, tmpfile.name)
    return tmpfile



from __future__ import annotations

import pytest

import daft
from daft.exceptions import ConnectTimeoutError, ReadTimeoutError


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_custom_s3fs(small_images_s3_paths):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)
    df = df.with_column(
        "data", df["urls"].url.download(io_config=daft.io.IOConfig(s3=daft.io.S3Config(anonymous=True)))
    )

    data = df.to_pydict()
    assert len(data["data"]) == 12
    for img_bytes in data["data"]:
        assert img_bytes is not None


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_custom_s3fs_wrong_region(small_images_s3_paths):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)
    df = df.with_column(
        "data", df["urls"].url.download(io_config=daft.io.IOConfig(s3=daft.io.S3Config(anonymous=True)))
    )

    data = df.to_pydict()
    assert len(data["data"]) == 12
    for img_bytes in data["data"]:
        assert img_bytes is not None


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_native_downloader(aws_public_s3_config, small_images_s3_paths):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(io_config=aws_public_s3_config))

    data = df.to_pydict()
    assert len(data["data"]) == 12
    for img_bytes in data["data"]:
        assert img_bytes is not None


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_native_downloader_with_connect_timeout(small_images_s3_paths):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)

    connect_timeout_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=True,
            connect_timeout_ms=1,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match="timed out when trying to connect to"):
        df = df.with_column("data", df["urls"].url.download(io_config=connect_timeout_config)).collect()


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_native_downloader_with_read_timeout(small_images_s3_paths):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)

    read_timeout_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=True,
            read_timeout_ms=1,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match="timed out when trying to connect to"):
        df = df.with_column("data", df["urls"].url.download(io_config=read_timeout_config)).collect()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Literal, cast

from daft.arrow_utils import ensure_table
from daft.daft import (
    CsvConvertOptions,
    CsvParseOptions,
    CsvReadOptions,
    JoinType,
    JsonConvertOptions,
    JsonParseOptions,
    JsonReadOptions,
    PySeries,
)
from daft.daft import PyRecordBatch as _PyRecordBatch
from daft.daft import ScanTask as _ScanTask
from daft.daft import read_csv as _read_csv
from daft.daft import read_json as _read_json
from daft.daft import read_parquet as _read_parquet
from daft.daft import read_parquet_bulk as _read_parquet_bulk
from daft.daft import read_parquet_into_pyarrow as _read_parquet_into_pyarrow
from daft.daft import read_parquet_into_pyarrow_bulk as _read_parquet_into_pyarrow_bulk
from daft.daft import read_parquet_statistics as _read_parquet_statistics
from daft.datatype import DataType, TimeUnit
from daft.dependencies import pa, pd
from daft.expressions import Expression, ExpressionsProjection
from daft.logical.schema import Schema
from daft.series import Series, item_to_series

if TYPE_CHECKING:
    from collections.abc import Mapping

    from daft.io import IOConfig

logger = logging.getLogger(__name__)


class RecordBatch:
    _recordbatch: _PyRecordBatch

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a RecordBatch via __init__ ")

    def schema(self) -> Schema:
        return Schema._from_pyschema(self._recordbatch.schema())

    def get_column(self, idx: int) -> Series:
        return Series._from_pyseries(self._recordbatch.get_column(idx))

    def columns(self) -> list[Series]:
        return [Series._from_pyseries(s) for s in self._recordbatch.columns()]

    def size_bytes(self) -> int:
        return self._recordbatch.size_bytes()

    def __len__(self) -> int:
        return len(self._recordbatch)

    def __repr__(self) -> str:
        return repr(self._recordbatch)

    def _repr_html_(self) -> str:
        return self._recordbatch._repr_html_()

    ###
    # Creation methods
    ###

    @staticmethod
    def empty(schema: Schema | None = None) -> RecordBatch:
        pyt = _PyRecordBatch.empty(None) if schema is None else _PyRecordBatch.empty(schema._schema)
        return RecordBatch._from_pyrecordbatch(pyt)

    @staticmethod
    def _from_scan_task(_: _ScanTask) -> RecordBatch:
        raise NotImplementedError("_from_scan_task is not implemented for legacy Python RecordBatch.")

    @staticmethod
    def _from_pyrecordbatch(pyt: _PyRecordBatch) -> RecordBatch:
        assert isinstance(pyt, _PyRecordBatch)
        tab = RecordBatch.__new__(RecordBatch)
        tab._recordbatch = pyt
        return tab

    @staticmethod
    def from_arrow_table(arrow_table: pa.Table) -> RecordBatch:
        assert isinstance(arrow_table, pa.Table)
        schema = Schema._from_field_name_and_types(
            [(f.name, DataType.from_arrow_type(f.type)) for f in arrow_table.schema]
        )
        non_native_fields = [
            field.name
            for field in schema
            if field.dtype == DataType.python()
            or field.dtype.is_tensor()
            or field.dtype.is_fixed_shape_tensor()
            or field.dtype.is_sparse_tensor()
            or field.dtype.is_fixed_shape_sparse_tensor()
        ]
        if non_native_fields:
            # If there are any contained Arrow types that are not natively supported, go through Table.from_pydict()
            # path.
            logger.debug("Unsupported Arrow types detected for columns: %s", non_native_fields)
            return RecordBatch.from_pydict(dict(zip(arrow_table.column_names, arrow_table.columns)))
        else:
            # Otherwise, go through record batch happy path.
            arrow_table = ensure_table(arrow_table)
            pyt = _PyRecordBatch.from_arrow_record_batches(arrow_table.to_batches(), schema._schema)
            return RecordBatch._from_pyrecordbatch(pyt)

    @staticmethod
    def from_arrow_record_batches(rbs: list[pa.RecordBatch], arrow_schema: pa.Schema) -> RecordBatch:
        schema = Schema._from_field_name_and_types([(f.name, DataType.from_arrow_type(f.type)) for f in arrow_schema])
        pyt = _PyRecordBatch.from_arrow_record_batches(rbs, schema._schema)
        return RecordBatch._from_pyrecordbatch(pyt)

    @staticmethod
    def from_pandas(pd_df: pd.DataFrame) -> RecordBatch:
        if not pd.module_available():  # type: ignore[attr-defined]
            raise ImportError("Unable to import Pandas - please ensure that it is installed.")
        assert isinstance(pd_df, pd.DataFrame)
        try:
            arrow_table = pa.Table.from_pandas(pd_df)
        except pa.ArrowInvalid:
            pass
        else:
            return RecordBatch.from_arrow_table(arrow_table)
        # Fall back to pydict path.
        df_as_dict = pd_df.to_dict(orient="series")
        if any(not isinstance(k, str) for k in df_as_dict.keys()):
            raise ValueError("pandas.DataFrame column names must be of type `str` to convert to a daft.RecordBatch.")
        df_as_dict_str = cast("dict[str, Any]", df_as_dict)
        return RecordBatch.from_pydict(df_as_dict_str)

    @staticmethod
    def from_pydict(data: Mapping[str, Any]) -> RecordBatch:
        series_dict: dict[str, PySeries] = dict()
        for k, v in data.items():
            series = item_to_series(k, v)
            series_dict[k] = series._series
        return RecordBatch._from_pyrecordbatch(_PyRecordBatch.from_pylist_series(series_dict))

    @classmethod
    def concat(cls, to_merge: list[RecordBatch]) -> RecordBatch:
        tables = []
        for t in to_merge:
            if not isinstance(t, RecordBatch):
                raise TypeError(f"Expected a Table for concat, got {type(t)}")
            tables.append(t._recordbatch)
        return RecordBatch._from_pyrecordbatch(_PyRecordBatch.concat(tables))

    def slice(self, start: int, end: int) -> RecordBatch:
        if not isinstance(start, int):
            raise TypeError(f"expected int for start but got {type(start)}")
        if not isinstance(end, int):
            raise TypeError(f"expected int for end but got {type(end)}")
        return RecordBatch._from_pyrecordbatch(self._recordbatch.slice(start, end))

    ###
    # Exporting methods
    ###

    def to_table(self) -> RecordBatch:
        """For compatibility with MicroPartition."""
        return self

    def to_arrow_record_batch(self) -> pa.RecordBatch:
        tab = pa.RecordBatch.from_pydict({column.name(): column.to_arrow() for column in self.columns()})
        return tab

    def to_arrow_table(self) -> pa.Table:
        tab = pa.Table.from_pydict({column.name(): column.to_arrow() for column in self.columns()})
        return tab

    def to_pydict(self) -> dict[str, list[Any]]:
        return {column.name(): column.to_pylist() for column in self.columns()}

    def to_pylist(self) -> list[dict[str, Any]]:
        # TODO(Clark): Avoid a double-materialization of the table once the Rust-side table supports
        # by-row selection or iteration.
        table = self.to_pydict()
        column_names = table.keys()
        return [{colname: table[colname][i] for colname in column_names} for i in range(len(self))]

    def to_pandas(
        self,
        schema: Schema | None = None,
        coerce_temporal_nanoseconds: bool = False,
    ) -> pd.DataFrame:
        from packaging.version import parse

        if not pd.module_available():  # type: ignore[attr-defined]
            raise ImportError("Unable to import Pandas - please ensure that it is installed.")

        python_fields = set()
        tensor_fields = set()
        for field in self.schema():
            if field.dtype.is_python():
                python_fields.add(field.name)
            elif field.dtype.is_tensor() or field.dtype.is_fixed_shape_tensor():
                tensor_fields.add(field.name)

        if python_fields or tensor_fields:
            table = {}
            for column_series in self.columns():
                colname = column_series.name()
                # Use Python list representation for Python typed columns or tensor columns (return as numpy)
                if colname in python_fields or colname in tensor_fields:
                    column = column_series.to_pylist()
                else:
                    # Arrow-native field, so provide column as Arrow array.
                    column_arrow = column_series.to_arrow()
                    if parse(pa.__version__) < parse("13.0.0"):
                        column = column_arrow.to_pandas()
                    else:
                        column = column_arrow.to_pandas(coerce_temporal_nanoseconds=coerce_temporal_nanoseconds)
                table[colname] = column

            return pd.DataFrame.from_dict(table)
        else:
            arrow_table = self.to_arrow_record_batch()
            if parse(pa.__version__) < parse("13.0.0"):
                return arrow_table.to_pandas()
            else:
                return arrow_table.to_pandas(coerce_temporal_nanoseconds=coerce_temporal_nanoseconds)

    ###
    # Compute methods (Table -> Table)
    ###

    def eval_expression_list(self, exprs: ExpressionsProjection) -> RecordBatch:
        assert all(isinstance(e, Expression) for e in exprs)
        pyexprs = [e._expr for e in exprs]
        return RecordBatch._from_pyrecordbatch(self._recordbatch.eval_expression_list(pyexprs))

    def head(self, num: int) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(self._recordbatch.head(num))

    def take(self, indices: Series) -> RecordBatch:
        assert isinstance(indices, Series)
        return RecordBatch._from_pyrecordbatch(self._recordbatch.take(indices._series))

    def filter(self, exprs: ExpressionsProjection) -> RecordBatch:
        assert all(isinstance(e, Expression) for e in exprs)
        pyexprs = [e._expr for e in exprs]
        return RecordBatch._from_pyrecordbatch(self._recordbatch.filter(pyexprs))

    def sort(
        self,
        sort_keys: ExpressionsProjection,
        descending: bool | list[bool] | None = None,
        nulls_first: bool | list[bool] | None = None,
    ) -> RecordBatch:
        assert all(isinstance(e, Expression) for e in sort_keys)
        pyexprs = [e._expr for e in sort_keys]
        if descending is None:
            descending = [False for _ in pyexprs]
        elif isinstance(descending, bool):
            descending = [descending for _ in pyexprs]
        elif isinstance(descending, list):
            if len(descending) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `descending` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(descending)} instead of {len(sort_keys)}"
                )
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending
        elif isinstance(nulls_first, bool):
            nulls_first = [nulls_first for _ in pyexprs]
        elif isinstance(nulls_first, list):
            if len(nulls_first) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
                )
            else:
                nulls_first = [bool(x) for x in nulls_first]
        return RecordBatch._from_pyrecordbatch(self._recordbatch.sort(pyexprs, descending, nulls_first))

    def sample(
        self,
        fraction: float | None = None,
        size: int | None = None,
        with_replacement: bool = False,
        seed: int | None = None,
    ) -> RecordBatch:
        if fraction is not None and size is not None:
            raise ValueError("Must specify either `fraction` or `size`, but not both")
        elif fraction is not None:
            return RecordBatch._from_pyrecordbatch(
                self._recordbatch.sample_by_fraction(fraction, with_replacement, seed)
            )
        elif size is not None:
            return RecordBatch._from_pyrecordbatch(self._recordbatch.sample_by_size(size, with_replacement, seed))
        else:
            raise ValueError("Must specify either `fraction` or `size`")

    def agg(self, to_agg: list[Expression], group_by: ExpressionsProjection | None = None) -> RecordBatch:
        to_agg_pyexprs = [e._expr for e in to_agg]
        group_by_pyexprs = [e._expr for e in group_by] if group_by is not None else []
        return RecordBatch._from_pyrecordbatch(self._recordbatch.agg(to_agg_pyexprs, group_by_pyexprs))

    def pivot(
        self,
        group_by: ExpressionsProjection,
        pivot_column: Expression,
        values_column: Expression,
        names: list[str],
    ) -> RecordBatch:
        group_by_pyexprs = [e._expr for e in group_by]
        return RecordBatch._from_pyrecordbatch(
            self._recordbatch.pivot(group_by_pyexprs, pivot_column._expr, values_column._expr, names)
        )

    def quantiles(self, num: int) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(self._recordbatch.quantiles(num))

    def explode(self, columns: ExpressionsProjection) -> RecordBatch:
        """NOTE: Expressions here must be Explode expressions."""
        to_explode_pyexprs = [e._expr for e in columns]
        return RecordBatch._from_pyrecordbatch(self._recordbatch.explode(to_explode_pyexprs))

    def hash_join(
        self,
        right: RecordBatch,
        left_on: ExpressionsProjection,
        right_on: ExpressionsProjection,
        how: JoinType = JoinType.Inner,
    ) -> RecordBatch:
        if len(left_on) != len(right_on):
            raise ValueError(
                f"Mismatch of number of join keys, left_on: {len(left_on)}, right_on: {len(right_on)}\nleft_on {left_on}\nright_on {right_on}"
            )

        if not isinstance(right, RecordBatch):
            raise TypeError(f"Expected a Table for `right` in join but got {type(right)}")

        left_exprs = [e._expr for e in left_on]
        right_exprs = [e._expr for e in right_on]

        return RecordBatch._from_pyrecordbatch(
            self._recordbatch.hash_join(right._recordbatch, left_on=left_exprs, right_on=right_exprs, how=how)
        )

    def sort_merge_join(
        self,
        right: RecordBatch,
        left_on: ExpressionsProjection,
        right_on: ExpressionsProjection,
        how: JoinType = JoinType.Inner,
        is_sorted: bool = False,
    ) -> RecordBatch:
        if how != JoinType.Inner:
            raise NotImplementedError("TODO: [RUST] Implement Other Join types")
        if len(left_on) != len(right_on):
            raise ValueError(
                f"Mismatch of number of join keys, left_on: {len(left_on)}, right_on: {len(right_on)}\nleft_on {left_on}\nright_on {right_on}"
            )

        if not isinstance(right, RecordBatch):
            raise TypeError(f"Expected a Table for `right` in join but got {type(right)}")

        left_exprs = [e._expr for e in left_on]
        right_exprs = [e._expr for e in right_on]

        return RecordBatch._from_pyrecordbatch(
            self._recordbatch.sort_merge_join(
                right._recordbatch,
                left_on=left_exprs,
                right_on=right_exprs,
                is_sorted=is_sorted,
            )
        )

    def partition_by_hash(self, exprs: ExpressionsProjection, num_partitions: int) -> list[RecordBatch]:
        if not isinstance(num_partitions, int):
            raise TypeError(f"Expected a num_partitions to be int, got {type(num_partitions)}")

        pyexprs = [e._expr for e in exprs]
        return [
            RecordBatch._from_pyrecordbatch(t) for t in self._recordbatch.partition_by_hash(pyexprs, num_partitions)
        ]

    def partition_by_range(
        self,
        partition_keys: ExpressionsProjection,
        boundaries: RecordBatch,
        descending: list[bool],
    ) -> list[RecordBatch]:
        if not isinstance(boundaries, RecordBatch):
            raise TypeError(f"Expected a RecordBatch for `boundaries` in partition_by_range but got {type(boundaries)}")

        exprs = [e._expr for e in partition_keys]
        return [
            RecordBatch._from_pyrecordbatch(t)
            for t in self._recordbatch.partition_by_range(exprs, boundaries._recordbatch, descending)
        ]

    def partition_by_random(self, num_partitions: int, seed: int) -> list[RecordBatch]:
        if not isinstance(num_partitions, int):
            raise TypeError(f"Expected a num_partitions to be int, got {type(num_partitions)}")

        if not isinstance(seed, int):
            raise TypeError(f"Expected a seed to be int, got {type(seed)}")

        return [RecordBatch._from_pyrecordbatch(t) for t in self._recordbatch.partition_by_random(num_partitions, seed)]

    def partition_by_value(self, partition_keys: ExpressionsProjection) -> tuple[list[RecordBatch], RecordBatch]:
        exprs = [e._expr for e in partition_keys]
        PyRecordBatchs, values = self._recordbatch.partition_by_value(exprs)

        return [RecordBatch._from_pyrecordbatch(t) for t in PyRecordBatchs], RecordBatch._from_pyrecordbatch(values)

    def add_monotonically_increasing_id(self, partition_num: int, column_name: str) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(
            self._recordbatch.add_monotonically_increasing_id(partition_num, column_name)
        )

    ###
    # Compute methods (Table -> Series)
    ###

    def argsort(
        self,
        sort_keys: ExpressionsProjection,
        descending: bool | list[bool] | None = None,
        nulls_first: bool | list[bool] | None = None,
    ) -> Series:
        assert all(isinstance(e, Expression) for e in sort_keys)
        pyexprs = [e._expr for e in sort_keys]
        if descending is None:
            descending = [False for _ in pyexprs]
        elif isinstance(descending, bool):
            descending = [descending for _ in pyexprs]
        elif isinstance(descending, list):
            if len(descending) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `descending` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(descending)} instead of {len(sort_keys)}"
                )
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending
        elif isinstance(nulls_first, bool):
            nulls_first = [nulls_first for _ in pyexprs]
        elif isinstance(nulls_first, list):
            if len(nulls_first) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
                )
            else:
                nulls_first = [bool(x) for x in nulls_first]
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
        return Series._from_pyseries(self._recordbatch.argsort(pyexprs, descending, nulls_first))

    def __reduce__(
        self,
    ) -> tuple[Callable[[dict[str, Any]], RecordBatch], tuple[dict[str, Series]]]:
        return RecordBatch.from_pydict, ({column.name(): column for column in self.columns()},)

    @classmethod
    def read_parquet(
        cls,
        path: str,
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups: list[int] | None = None,
        predicate: Expression | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    ) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(
            _read_parquet(
                uri=path,
                columns=columns,
                start_offset=start_offset,
                num_rows=num_rows,
                row_groups=row_groups,
                predicate=predicate._expr if predicate is not None else None,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
                coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
            )
        )

    @classmethod
    def read_parquet_bulk(
        cls,
        paths: list[str],
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups_per_path: list[list[int] | None] | None = None,
        predicate: Expression | None = None,
        io_config: IOConfig | None = None,
        num_parallel_tasks: int | None = 128,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    ) -> list[RecordBatch]:
        PyRecordBatchs = _read_parquet_bulk(
            uris=paths,
            columns=columns,
            start_offset=start_offset,
            num_rows=num_rows,
            row_groups=row_groups_per_path,
            predicate=predicate._expr if predicate is not None else None,
            io_config=io_config,
            num_parallel_tasks=num_parallel_tasks,
            multithreaded_io=multithreaded_io,
            coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
        )
        return [RecordBatch._from_pyrecordbatch(t) for t in PyRecordBatchs]

    @classmethod
    def read_parquet_statistics(
        cls,
        paths: Series | list[str],
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> RecordBatch:
        if not isinstance(paths, Series):
            paths = Series.from_pylist(paths, name="uris").cast(DataType.string())
        assert paths.name() == "uris", f"Expected input series to have name 'uris', but found: {paths.name()}"
        return RecordBatch._from_pyrecordbatch(
            _read_parquet_statistics(
                uris=paths._series,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )

    @classmethod
    def read_csv(
        cls,
        path: str,
        convert_options: CsvConvertOptions | None = None,
        parse_options: CsvParseOptions | None = None,
        read_options: CsvReadOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(
            _read_csv(
                uri=path,
                convert_options=convert_options,
                parse_options=parse_options,
                read_options=read_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )

    @classmethod
    def read_json(
        cls,
        path: str,
        convert_options: JsonConvertOptions | None = None,
        parse_options: JsonParseOptions | None = None,
        read_options: JsonReadOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
        max_chunks_in_flight: int | None = None,
    ) -> RecordBatch:
        return RecordBatch._from_pyrecordbatch(
            _read_json(
                uri=path,
                convert_options=convert_options,
                parse_options=parse_options,
                read_options=read_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
                max_chunks_in_flight=max_chunks_in_flight,
            )
        )


def read_parquet_into_pyarrow(
    path: str,
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups: list[int] | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    string_encoding: Literal["utf-8", "raw"] = "utf-8",
    file_timeout_ms: int | None = 900_000,  # 15 minutes
) -> pa.Table:
    fields, metadata, columns, num_rows_read = _read_parquet_into_pyarrow(
        uri=path,
        columns=columns,
        start_offset=start_offset,
        num_rows=num_rows,
        row_groups=row_groups,
        io_config=io_config,
        multithreaded_io=multithreaded_io,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
        string_encoding=string_encoding,
        file_timeout_ms=file_timeout_ms,
    )
    schema = pa.schema(fields, metadata=metadata)
    columns = [pa.chunked_array(c, type=f.type) for f, c in zip(schema, columns)]  # type: ignore

    if columns:
        return pa.table(columns, schema=schema)
    else:
        # If data contains no columns, we return an empty table with the appropriate size using `Table.drop`
        return pa.table({"dummy_column": pa.array([None] * num_rows_read)}).drop(["dummy_column"])


def read_parquet_into_pyarrow_bulk(
    paths: list[str],
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups_per_path: list[list[int] | None] | None = None,
    io_config: IOConfig | None = None,
    num_parallel_tasks: int | None = 128,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
) -> list[pa.Table]:
    bulk_result = _read_parquet_into_pyarrow_bulk(
        uris=paths,
        columns=columns,
        start_offset=start_offset,
        num_rows=num_rows,
        row_groups=row_groups_per_path,
        io_config=io_config,
        num_parallel_tasks=num_parallel_tasks,
        multithreaded_io=multithreaded_io,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
    )

    tables = []
    for fields, metadata, columns, num_rows_read in bulk_result:
        if columns:
            table = pa.table(
                [pa.chunked_array(c, type=f.type) for f, c in zip(fields, columns)],
                schema=pa.schema(fields, metadata=metadata),
            )
        else:
            # If data contains no columns, we return an empty table with the appropriate size using `Table.drop`
            table = pa.table({"dummy_col": [None] * num_rows_read}).drop(["dummy_col"])
        tables.append(table)
    return tables



from __future__ import annotations

import ast

import pytest

import daft
from daft.sql import SQLCatalog


def test_sql_udf():
    df = daft.from_pydict({"a": [1, 2, 3]})

    @daft.udf(return_dtype=daft.DataType.int64())
    def multiply_by_n(data, n):
        return [i * n for i in data]

    cat = SQLCatalog({"df": df})

    expected = {"b": [2, 4, 6]}
    actual = daft.sql("select multiply_by_n(a, n:=2) as b from df", cat).to_pydict()

    assert actual == expected


def test_sql_udf_ambigious_name():
    df = daft.from_pydict({"a": [1, 2, 3]})
    cat = SQLCatalog({"df": df})

    @daft.udf(return_dtype=daft.DataType.int64())
    def multiply_by_n(data, n):
        return [i * n for i in data]

    @daft.udf(return_dtype=daft.DataType.int64())
    def MULTIPLY_BY_N(data, n):
        return [i * n for i in data]

    with pytest.raises(Exception, match="Invalid operation: Ambigiuous identifier for Function: found"):
        daft.sql("select multiply_by_n(a, n:=2) from df", cat).to_pydict()


def test_sql_udf_multi_column():
    df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
    cat = SQLCatalog({"df": df})

    @daft.udf(return_dtype=daft.DataType.int64())
    def multiply(a, b):
        return a * b

    expected = {"a": [4, 10, 18]}
    actual = daft.sql("select multiply(a, b) as a from df", cat).to_pydict()
    assert actual == expected


def test_sql_udf_multi_column_and_kwargs():
    df = daft.from_pydict({"first_name": ["Alice", "Bob", "Charlie"], "last_name": ["Smith", "Johnson", "Williams"]})
    cat = SQLCatalog({"df": df})

    @daft.udf(return_dtype=str)
    def make_greeting(a, b, greeting="hello"):
        return [f"{greeting}, {a} {b}" for a, b in zip(a, b)]

    expected = {
        "greeting1": ["hello, Alice Smith", "hello, Bob Johnson", "hello, Charlie Williams"],
        "greeting2": ["hi, Alice Smith", "hi, Bob Johnson", "hi, Charlie Williams"],
    }
    actual = daft.sql(
        """
        select
            make_greeting(first_name, last_name) as greeting1,
            make_greeting(first_name, last_name, greeting:='hi') as greeting2
        from df
    """,
        cat,
    ).to_pydict()
    assert actual == expected


# to differentiate between literals & arguments, the scalar arguments to udfs need to be passed as keyword arguments
@pytest.mark.skip(reason="doesn't work as expected because udf's don error if the args are incorrect")
def test_sql_udf_kwargs_dont_work_as_positional():
    pass


@pytest.mark.parametrize(
    "value,return_type",
    [("1.23", float), ("1", int), ("'hello'", str), ("['hello']", daft.DataType.list(daft.DataType.string()))],
)
def test_sql_udf_various_datatypes(value, return_type):
    @daft.udf(return_dtype=return_type)
    def udf(column, literal):
        return [literal for i in column]

    df = daft.from_pydict({"x": [i for i in range(10)]})

    cat = SQLCatalog({"df": df})
    query = f"select udf(x, literal:={value}) as x from df"
    actual = daft.sql(query, cat).to_pydict()

    if return_type is str:
        # remove the quotes
        value = value[1:-1]
    elif return_type is float:
        value = float(value)
    elif return_type is int:
        value = int(value)
    elif return_type == daft.DataType.list(daft.DataType.string()):
        value = ast.literal_eval(value)
    expected = {"x": [value] * 10}
    assert actual == expected


def test_sql_udf_unregister():
    @daft.udf(return_dtype=str)
    def my_udf(column, literal):
        return ["hello" for i in column]

    daft.detach_function("my_udf")
    with pytest.raises(Exception, match="Function `my_udf` not found"):
        daft.sql("select my_udf(1, literal:='world')").to_pydict()



from __future__ import annotations


def test_pipe(make_df):
    df = make_df({"x": [1, 2, 3]})

    def double(df, column: str):
        return df.select((df[column] * df[column]).alias("y"))

    assert df.pipe(double, "x").to_pydict() == {"y": [1, 4, 9]}


def test_pipe_with_join(make_df):
    df1 = make_df({"A": [1, 2, 3], "B": [1, 2, 3]})
    df2 = make_df({"A": [1, 2, 3], "C": [1, 2, 3]})

    def left_join(dfa, dfb, on):
        return dfa.join(dfb, on=on)

    joined_by_join = df1.join(df2, on="A")
    joined_by_pipe = df1.pipe(left_join, df2, on="A")

    assert joined_by_join.to_pydict() == joined_by_pipe.to_pydict()



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    ["expr", "data", "expected"],
    [
        (col("col").str.split(","), ["a,b,c", "d,e", "f", "g,h"], [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]]),
        (
            col("col").str.split(lit(",")),
            ["a,b,c", "d,e", "f", "g,h"],
            [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]],
        ),
        (
            col("col").str.split(col("emptystrings") + lit(",")),
            ["a,b,c", "d,e", "f", "g,h"],
            [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]],
        ),
        (
            col("col").str.split(r",+", True),
            ["a,,,,b,,,,c", "d,,,e", "f", "g,h"],
            [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]],
        ),
        (
            col("col").str.split(lit(r",+"), True),
            ["a,,,,b,,,,c", "d,,,e", "f", "g,h"],
            [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]],
        ),
        (
            col("col").str.split(col("emptystrings") + lit(r",+"), True),
            ["a,,,,b,,,,c", "d,,,e", "f", "g,h"],
            [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]],
        ),
    ],
)
def test_series_utf8_split_broadcast_pattern(expr, data, expected) -> None:
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", "", "", ""]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": expected}



from __future__ import annotations

import copy

import pyarrow as pa
import pytest

from daft.logical.schema import Schema
from daft.recordbatch.micropartition import MicroPartition


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": pa.array([1, 2, 3], type=pa.int64())}),
        MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64()}))),  # No tables
    ],
)
def test_pickling_loaded(mp) -> None:
    assert mp.to_arrow() == copy.deepcopy(mp).to_arrow()


def test_pickling_unloaded() -> None:
    mp = MicroPartition.read_parquet("tests/assets/parquet-data/parquet-with-schema-metadata.parquet")
    assert copy.deepcopy(mp).to_arrow() == mp.to_arrow()



from __future__ import annotations

import itertools

import pytest
import s3fs

from daft.daft import io_glob

from ..conftest import minio_create_bucket

NUM_FILES = 10000
NUM_LEVELS = 4
FANOUT_PER_LEVEL = 12
BUCKET = "bucket"


def generate_one_file_per_dir():
    # Total of 10k files (10^4 * 1)
    NUM_LEVELS = 4
    FANOUT_PER_LEVEL = 10
    return [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate(part_vals)) + "/0.parquet"
        for part_vals in itertools.product([str(i) for i in range(FANOUT_PER_LEVEL)], repeat=NUM_LEVELS)
    ]


def generate_balanced_partitioned_data():
    # Total of 10k files (10^3 * 10)
    NUM_LEVELS = 3
    FANOUT_PER_LEVEL = 10
    FILES_PER_PARTITION = 10
    return [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate(part_vals)) + f"/{i}.parquet"
        for i in range(FILES_PER_PARTITION)
        for part_vals in itertools.product([str(i) for i in range(FANOUT_PER_LEVEL)], repeat=NUM_LEVELS)
    ]


def generate_left_skew_partitioned_data():
    # Total of 10k files (10^3 * 10)
    NUM_LEVELS = 3
    FANOUT_PER_LEVEL = 10

    # First partition contains 1009 files, and the other partitions have 9 files each
    num_files_per_partition = [1009] + [9 for i in range(999)]

    return [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate(part_vals)) + f"/{i}.parquet"
        for part_vals, num_files_for_this_partition in zip(
            itertools.product([str(i) for i in range(FANOUT_PER_LEVEL)], repeat=NUM_LEVELS), num_files_per_partition
        )
        for i in range(num_files_for_this_partition)
    ]


def generate_right_skew_partitioned_data():
    # Total of 10k files (10^3 * 10)
    NUM_LEVELS = 3
    FANOUT_PER_LEVEL = 10

    # Last partition contains 1009 files, and the other partitions have 9 files each
    num_files_per_partition = [9 for i in range(999)] + [1009]

    return [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate(part_vals)) + f"/{i}.parquet"
        for part_vals, num_files_for_this_partition in zip(
            itertools.product([str(i) for i in range(FANOUT_PER_LEVEL)], repeat=NUM_LEVELS), num_files_per_partition
        )
        for i in range(num_files_for_this_partition)
    ]


def generate_left_skew_dirs_partitioned_data():
    # First partition in level 0 fans out (1 * 20 * 20 * 10 files = 4000 files)
    # The rest of the partitions in level 0 fan out much smaller (1 * 8 * 8 * 10 = 10 files) * 9 = 5940 files
    first_partition_paths = [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate([0, *part_vals])) + f"/{i}.parquet"
        for i in range(10)
        for part_vals in itertools.product([str(i) for i in range(20)], repeat=2)
    ]
    other_partition_paths = [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate([first_level_val, *part_vals])) + f"/{i}.parquet"
        for first_level_val in range(1, 10)
        for part_vals in itertools.product([str(i) for i in range(8)], repeat=2)
        for i in range(10)
    ]

    return first_partition_paths + other_partition_paths


def generate_right_skew_dirs_partitioned_data():
    # Last partition in level 0 fans out (1 * 20 * 20 * 10 files = 4000 files)
    # The rest of the partitions in level 0 fan out much smaller (1 * 8 * 8 * 10 = 10 files) * 9 = 5940 files
    last_partition_paths = [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate([9, *part_vals])) + f"/{i}.parquet"
        for i in range(10)
        for part_vals in itertools.product([str(i) for i in range(20)], repeat=2)
    ]
    other_partition_paths = [
        "/".join(f"part_col_{i}={val}" for i, val in enumerate([first_level_val, *part_vals])) + f"/{i}.parquet"
        for first_level_val in range(0, 9)
        for part_vals in itertools.product([str(i) for i in range(8)], repeat=2)
        for i in range(10)
    ]

    return other_partition_paths + last_partition_paths


def generate_bushy_late_partitioned_data():
    # Total of 10k files (10^3 * 10)
    return [f"single/single/part_col={val}" + f"/{i}.parquet" for i in range(10) for val in range(1000)]


def generate_bushy_early_partitioned_data():
    # Total of 10k files (10^3 * 10)
    return [f"part_col={val}/single/single" + f"/{i}.parquet" for i in range(10) for val in range(1000)]


FILE_NAME_GENERATORS = {
    "one-file-per-dir": generate_one_file_per_dir,
    "partitioned-data-balanced": generate_balanced_partitioned_data,
    "partitioned-data-left-skew-files": generate_left_skew_partitioned_data,
    "partitioned-data-right-skew-files": generate_right_skew_partitioned_data,
    "partitioned-data-left-skew-dirs": generate_left_skew_dirs_partitioned_data,
    "partitioned-data-right-skew-dirs": generate_right_skew_dirs_partitioned_data,
    "partitioned-data-bushy-early": generate_bushy_early_partitioned_data,
    "partitioned-data-bushy-late": generate_bushy_late_partitioned_data,
}


@pytest.fixture(
    scope="module",
    params=[
        "one-file-per-dir",
        "partitioned-data-balanced",
        "partitioned-data-left-skew-files",
        "partitioned-data-right-skew-files",
        "partitioned-data-right-skew-dirs",
        "partitioned-data-left-skew-dirs",
        "partitioned-data-bushy-early",
        "partitioned-data-bushy-late",
    ],
)
def setup_bucket(request, minio_io_config):
    test_name = request.param
    file_name_generator = FILE_NAME_GENERATORS[test_name]
    with minio_create_bucket(minio_io_config, bucket_name=BUCKET) as fs:
        files = file_name_generator()
        print(f"Num files: {len(files)}")
        for name in files:
            fs.touch(f"{BUCKET}/{name}")
        yield len(set(files))


@pytest.mark.benchmark(group="glob")
@pytest.mark.integration()
def test_benchmark_glob_s3fs(benchmark, setup_bucket, minio_io_config):
    fs = s3fs.S3FileSystem(
        key=minio_io_config.s3.key_id,
        password=minio_io_config.s3.access_key,
        client_kwargs={"endpoint_url": minio_io_config.s3.endpoint_url},
    )
    results = benchmark(
        lambda: fs.glob(
            f"s3://{BUCKET}/**/*.parquet",
            # Can't set page size for s3fs
            # max_items=PAGE_SIZE,
        )
    )
    assert len(results) == setup_bucket


@pytest.mark.benchmark(group="glob")
@pytest.mark.integration()
@pytest.mark.parametrize("page_size", [100, 1000])
def test_benchmark_glob_boto3_list(benchmark, setup_bucket, minio_io_config, page_size):
    import boto3

    s3 = boto3.client(
        "s3",
        aws_access_key_id=minio_io_config.s3.key_id,
        aws_secret_access_key=minio_io_config.s3.access_key,
        endpoint_url=minio_io_config.s3.endpoint_url,
    )

    def f():
        continuation_token = None
        has_next = True
        opts = {"MaxKeys": page_size}
        data = []
        while has_next:
            if continuation_token is not None:
                opts["ContinuationToken"] = continuation_token

            response = s3.list_objects_v2(
                Bucket=BUCKET,
                Prefix="",
                **opts,
            )
            data.extend([{"path": f"s3://{BUCKET}/{d['Key']}", "type": "File"} for d in response.get("Contents", [])])
            continuation_token = response.get("NextContinuationToken")
            has_next = continuation_token is not None

        return data

    data = benchmark(f)
    assert len(data) == setup_bucket


@pytest.mark.benchmark(group="glob")
@pytest.mark.integration()
@pytest.mark.parametrize("fanout_limit", [128, 256])
@pytest.mark.parametrize("page_size", [100, 1000])
def test_benchmark_glob_daft(benchmark, setup_bucket, minio_io_config, fanout_limit, page_size):
    results = benchmark(
        lambda: io_glob(
            f"s3://{BUCKET}/**/*.parquet", io_config=minio_io_config, fanout_limit=fanout_limit, page_size=page_size
        )
    )
    assert len(results) == setup_bucket



# /// script
# requires-python = ">=3.12"
# dependencies = ["PyGithub"]
# ///

import argparse
import os
import pathlib

from github import Auth, Github, enable_console_debug_logging

GITHUB_OAUTH_TOKEN_ENVVAR = "GITHUB_OAUTH_TOKEN"


def get_oauth_token() -> str:
    token = os.getenv(GITHUB_OAUTH_TOKEN_ENVVAR)
    if token is None:
        raise RuntimeError(
            "Unable to find github token at $GITHUB_OAUTH_TOKEN. Please generate one and assign it to the environment variable. "
            "We recommend creating a Classic token (the fine-grained ones seem to not really work) and give it at least Workflow as well as Repository access. "
            "https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens"
        )
    return token


def main(
    daft_wheel_url: str | None,
    daft_version: str,
    path_to_script: pathlib.Path,
    script_args: str | None,
    branch: str = "main",
):
    auth = Auth.Token(get_oauth_token())
    g = Github(auth=auth)

    daft_package = {"daft_wheel_url": daft_wheel_url} if daft_wheel_url is not None else {"daft_version": daft_version}
    entrypoint_args = {"entrypoint_args": script_args} if script_args is not None else {}

    inputs = {
        **daft_package,
        "entrypoint_script": str(path_to_script.name),
        "working_dir": str(path_to_script.parent),
        **entrypoint_args,
    }

    repo = g.get_repo("Eventual-Inc/Daft")
    workflow = repo.get_workflow("run-cluster.yaml")

    created = workflow.create_dispatch(
        ref=branch,
        inputs={**inputs},
    )

    if not created:
        raise RuntimeError("Could not create workflow, suggestion: run again with --verbose")

    print("Workflow created, view it at: https://github.com/Eventual-Inc/Daft/actions/workflows/run-cluster.yaml")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Run Daft cluster workflow on github actions",
        usage="uv run gha_run_cluster_job.py [-h] [--branch BRANCH] [--daft-wheel-url DAFT_WHEEL_URL | --daft-version DAFT_VERSION] path_to_script [-- script_args ...]",
    )
    parser.add_argument("path_to_script", type=pathlib.Path, help="Path to the script to run")
    parser.add_argument("--branch", default="main", help="GitHub branch to use")
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("--daft-wheel-url", help="URL of the Daft wheel")
    group.add_argument("--daft-version", default="0.3.15", help="Released version of daft (default: 0.3.15)")
    parser.add_argument("--verbose", action="store_true", help="Enable verbose mode")
    args, extra = parser.parse_known_args()

    if args.verbose:
        enable_console_debug_logging()

    script_args = None
    if extra:
        if extra[0] != "--":
            raise ValueError(f"Arguments to the script must be passed in after a '--', received: {' '.join(extra)}")
        script_args = " ".join(extra[1:])

    main(
        daft_wheel_url=args.daft_wheel_url,
        daft_version=args.daft_version,
        path_to_script=args.path_to_script,
        script_args=script_args,
        branch=args.branch,
    )



from __future__ import annotations

import random
from uuid import uuid4

import numpy as np
import pytest

import daft
from daft import DataFrame, col


@pytest.mark.benchmark(group="sorts")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
def test_sort_simple(benchmark, num_samples, num_partitions) -> None:
    """Test simple sort performance.

    Keys are consecutive integers; no data payload.
    """
    arr = np.arange(num_samples)
    np.random.shuffle(arr)

    df = (
        daft.from_pydict(
            {
                "mykey": arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.sort("mykey").collect()

    result = benchmark(bench)

    # Make sure the result is correct.
    assert (result.to_pandas()["mykey"].to_numpy() == np.arange(num_samples)).all()


@pytest.mark.benchmark(group="sorts")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
def test_sort_strings(benchmark, num_samples, num_partitions) -> None:
    """Test the impact of string keys vs integer keys."""
    keys = [str(uuid4()) for _ in range(num_samples)]
    random.shuffle(keys)

    df = (
        daft.from_pydict(
            {
                "mykey": keys,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.sort("mykey").collect()

    result = benchmark(bench)

    # Make sure the result is correct.
    keys.sort()
    assert result.to_pydict()["mykey"] == keys


@pytest.mark.benchmark(group="sorts")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
def test_sort_withdata(benchmark, num_samples, num_partitions) -> None:
    """Test the impact of data payloads."""
    arr = np.arange(num_samples)
    np.random.shuffle(arr)

    long_A = "A" * 1024

    df = (
        daft.from_pydict(
            {
                "mykey": arr,
                "data": [long_A for _ in range(num_samples)],
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.sort("mykey").collect()

    result = benchmark(bench)

    # Make sure the result is correct.
    assert (result.to_pandas()["mykey"].to_numpy() == np.arange(num_samples)).all()
    assert result.groupby("data").agg(col("mykey").count()).to_pydict() == {
        "data": [long_A],
        "mykey": [num_samples],
    }


@pytest.mark.benchmark(group="sorts")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
@pytest.mark.parametrize("num_columns", [1, 4], ids=["1col", "4cols"])
def test_multicolumn_sort(benchmark, num_columns, num_samples, num_partitions) -> None:
    """Evaluate the performance impact of sorting multiple columns.

    Each additional sort column increases the sort key cardinality by approximately the same factor of ~10
    (i.e. each additional sort column should be doing around the same additional amount of work).

    Using all columns produces a unique sort key.
    """
    arr = np.arange(num_samples)
    np.random.shuffle(arr)

    df = (
        daft.from_pydict(
            {
                # all coprime
                "nums_9": arr * 17 % 9,
                "nums_10": arr * 17 % 10,
                "nums_11": arr * 17 % 11,
                "nums": arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    sort_on = ["nums_9", "nums_10", "nums_11", "nums"][:num_columns]

    def bench() -> DataFrame:
        return df.sort(sort_on).collect()

    benchmark(bench)



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import (
    CsvSourceConfig,
    FileFormatConfig,
    IOConfig,
    StorageConfig,
)
from daft.dataframe import DataFrame
from daft.datatype import DataType
from daft.io.common import get_tabular_files_scan


@PublicAPI
def read_csv(
    path: Union[str, list[str]],
    infer_schema: bool = True,
    schema: Optional[dict[str, DataType]] = None,
    has_headers: bool = True,
    delimiter: Optional[str] = None,
    double_quote: bool = True,
    quote: Optional[str] = None,
    escape_char: Optional[str] = None,
    comment: Optional[str] = None,
    allow_variable_columns: bool = False,
    io_config: Optional[IOConfig] = None,
    file_path_column: Optional[str] = None,
    hive_partitioning: bool = False,
    schema_hints: Optional[dict[str, DataType]] = None,
    _buffer_size: Optional[int] = None,
    _chunk_size: Optional[int] = None,
) -> DataFrame:
    """Creates a DataFrame from CSV file(s).

    Args:
        path (str): Path to CSV (allows for wildcards)
        infer_schema (bool): Whether to infer the schema of the CSV, defaults to True.
        schema (dict[str, DataType]): A schema that is used as the definitive schema for the CSV if infer_schema is False, otherwise it is used as a schema hint that is applied after the schema is inferred.
        has_headers (bool): Whether the CSV has a header or not, defaults to True
        delimiter (Str): Delimiter used in the CSV, defaults to ","
        doubled_quote (bool): Whether to support double quote escapes, defaults to True
        escape_char (str): Character to use as the escape character for double quotes, or defaults to `"`
        comment (str): Character to treat as the start of a comment line, or None to not support comments
        allow_variable_columns (bool): Whether to allow for variable number of columns in the CSV, defaults to False. If set to True, Daft will append nulls to rows with less columns than the schema, and ignore extra columns in rows with more columns
        io_config (IOConfig): Config to be used with the native downloader
        file_path_column: Include the source path(s) as a column with this name. Defaults to None.
        hive_partitioning: Whether to infer hive_style partitions from file paths and include them as columns in the Dataframe. Defaults to False.

    Returns:
        DataFrame: parsed DataFrame

    Examples:
        >>> df = daft.read_csv("/path/to/file.csv")
        >>> df = daft.read_csv("/path/to/directory")
        >>> df = daft.read_csv("/path/to/files-*.csv")
        >>> df = daft.read_csv("s3://path/to/files-*.csv")

    """
    if isinstance(path, list) and len(path) == 0:
        raise ValueError("Cannot read DataFrame from from empty list of CSV filepaths")

    if schema_hints is not None:
        raise ValueError(
            "Specifying schema_hints is deprecated from Daft version >= 0.3.0! Instead, please use the 'schema' and 'infer_schema' arguments."
        )

    if not infer_schema and schema is None:
        raise ValueError(
            "Cannot read DataFrame with infer_schema=False and schema=None, please provide a schema or set infer_schema=True"
        )

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config

    csv_config = CsvSourceConfig(
        delimiter=delimiter,
        has_headers=has_headers,
        double_quote=double_quote,
        quote=quote,
        escape_char=escape_char,
        comment=comment,
        allow_variable_columns=allow_variable_columns,
        buffer_size=_buffer_size,
        chunk_size=_chunk_size,
    )
    file_format_config = FileFormatConfig.from_csv_config(csv_config)
    storage_config = StorageConfig(True, io_config)

    builder = get_tabular_files_scan(
        path=path,
        infer_schema=infer_schema,
        schema=schema,
        file_format_config=file_format_config,
        storage_config=storage_config,
        file_path_column=file_path_column,
        hive_partitioning=hive_partitioning,
    )
    return DataFrame(builder)



from __future__ import annotations

import itertools
import operator
from datetime import date, datetime

import pyarrow as pa
import pytest
import pytz

from daft import DataType, Series

arrow_int_types = [
    pa.int8(),
    pa.uint8(),
    pa.int16(),
    pa.uint16(),
    pa.int32(),
    pa.uint32(),
    pa.int64(),
    pa.uint64(),
]
arrow_decimal_types = [pa.decimal128(20, 5), pa.decimal128(15, 9)]
arrow_string_types = [pa.string(), pa.large_string()]
arrow_float_types = [pa.float32(), pa.float64()]
arrow_binary_types = [pa.binary(), pa.large_binary()]
arrow_number_types = arrow_int_types + arrow_decimal_types + arrow_float_types


VALID_INT_STRING_COMPARISONS = list(itertools.product(arrow_int_types + arrow_decimal_types, repeat=2)) + list(
    itertools.product(arrow_string_types, repeat=2)
)


def make_array(data: list, type=None) -> pa.array:
    if type is not None and (pa.types.is_string(type) or pa.types.is_large_string(type)):
        data = [str(x) if x is not None else None for x in data]
    return pa.array(data, type=type)


@pytest.mark.parametrize("l_dtype, r_dtype", VALID_INT_STRING_COMPARISONS)
def test_comparisons_int_and_str(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([1, 3, 1, 5, None, None], type=r_dtype)
    # eq, lt, gt, None, None, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [False, True, False, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, None, None]


@pytest.mark.parametrize("l_dtype, r_dtype", VALID_INT_STRING_COMPARISONS)
def test_comparisons_int_and_str_left_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([2], type=l_dtype)
    r_arrow = make_array([1, 2, 3, None], type=r_dtype)
    # gt, eq, lt

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [False, False, True, None]

    le = (left <= right).to_pylist()
    assert le == [False, True, True, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, True, False, None]

    gt = (left > right).to_pylist()
    assert gt == [True, False, False, None]


@pytest.mark.parametrize("l_dtype, r_dtype", VALID_INT_STRING_COMPARISONS)
def test_comparisons_int_and_str_right_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([2], type=r_dtype)
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [True, False, False, None, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, False, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [False, True, True, None, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, True, None]


@pytest.mark.parametrize("l_dtype, r_dtype", VALID_INT_STRING_COMPARISONS)
def test_comparisons_int_and_str_right_null_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([None], type=r_dtype)
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [None, None, None, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [None, None, None, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [None, None, None, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [None, None, None, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [None, None, None, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [None, None, None, None, None, None]


@pytest.mark.parametrize(
    "l_dtype, r_dtype",
    itertools.product(arrow_int_types + arrow_float_types + arrow_decimal_types, repeat=2),
)
def test_comparisons_int_and_float(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([1, 3, 1, 5, None, None], type=r_dtype)
    # eq, lt, gt, None, None, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [False, True, False, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, None, None]


@pytest.mark.parametrize(
    "l_dtype, r_dtype",
    itertools.product(arrow_int_types + arrow_float_types + arrow_decimal_types, repeat=2),
)
def test_comparisons_int_and_float_right_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([2], type=r_dtype)
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [True, False, False, None, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, False, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [False, True, True, None, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, True, None]


@pytest.mark.parametrize(
    "l_dtype, r_dtype",
    itertools.product(arrow_int_types + arrow_float_types + arrow_decimal_types, repeat=2),
)
def test_comparisons_int_and_float_right_null_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None], type=l_dtype)
    r_arrow = make_array([None], type=r_dtype)
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [None, None, None, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [None, None, None, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [None, None, None, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [None, None, None, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [None, None, None, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [None, None, None, None, None, None]


def test_comparisons_boolean_array() -> None:
    l_arrow = make_array([True, False, True, False, True, False, None, None, None])
    r_arrow = make_array([True, True, False, False, None, None, True, False, None])
    # lt, eq, lt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [False, True, False, False, None, None, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, True, None, None, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, True, None, None, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, False, None, None, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, True, None, None, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, False, None, None, None, None, None]

    _and = (left & right).to_pylist()
    assert _and == [True, False, False, False, None, False, None, False, None]

    _or = (left | right).to_pylist()
    assert _or == [True, True, True, False, True, None, True, None, None]

    _xor = (left ^ right).to_pylist()
    assert _xor == [False, True, True, False, None, None, None, None, None]


def test_comparisons_boolean_array_right_scalar() -> None:
    l_arrow = make_array([False, True, None])
    r_arrow = make_array([True])

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [True, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, None]

    ge = (left >= right).to_pylist()
    assert ge == [False, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, None]

    _and = (left & right).to_pylist()
    assert _and == [False, True, None]

    _or = (left | right).to_pylist()
    assert _or == [True, True, True]

    _xor = (left ^ right).to_pylist()
    assert _xor == [True, False, None]

    r_arrow = make_array([False])
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [False, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, False, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, True, None]

    _and = (left & right).to_pylist()
    assert _and == [False, False, False]

    _or = (left | right).to_pylist()
    assert _or == [False, True, None]

    _xor = (left ^ right).to_pylist()
    assert _xor == [False, True, None]

    r_arrow = make_array([None], type=pa.bool_())
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [None, None, None]

    le = (left <= right).to_pylist()
    assert le == [None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [None, None, None]

    _and = (left & right).to_pylist()
    assert _and == [False, None, None]

    _or = (left | right).to_pylist()
    assert _or == [None, True, None]

    _xor = (left ^ right).to_pylist()
    assert _xor == [None, None, None]


def test_comparisons_boolean_array_left_scalar() -> None:
    l_arrow = make_array([True])
    r_arrow = make_array([False, True, None])
    # lt, eq, lt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [False, False, None]

    le = (left <= right).to_pylist()
    assert le == [False, True, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, True, None]

    gt = (left > right).to_pylist()
    assert gt == [True, False, None]

    _and = (left & right).to_pylist()
    assert _and == [False, True, None]

    _oright = (left | right).to_pylist()
    assert _oright == [True, True, True]

    _xoright = (left ^ right).to_pylist()
    assert _xoright == [True, False, None]


def test_comparisons_bad_right_value() -> None:
    l_arrow = make_array([1, 2, 3, None, 5, None])

    left = Series.from_arrow(l_arrow)
    right = [1, 2, 3, None, 5, None]

    with pytest.raises(TypeError, match="another Series"):
        left < right

    with pytest.raises(TypeError, match="another Series"):
        left <= right

    with pytest.raises(TypeError, match="another Series"):
        left == right

    with pytest.raises(TypeError, match="another Series"):
        left != right

    with pytest.raises(TypeError, match="another Series"):
        left >= right

    with pytest.raises(TypeError, match="another Series"):
        left > right

    with pytest.raises(TypeError, match="another Series"):
        left & right

    with pytest.raises(TypeError, match="another Series"):
        left | right

    with pytest.raises(TypeError, match="another Series"):
        left ^ right


def test_boolean_array_mismatch_length() -> None:
    l_arrow = make_array([False, True, None, None])
    r_arrow = make_array([False, True, False, True, None])

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    with pytest.raises(ValueError, match="different length"):
        left < right

    with pytest.raises(ValueError, match="different length"):
        left <= right

    with pytest.raises(ValueError, match="different length"):
        left == right

    with pytest.raises(ValueError, match="different length"):
        left != right

    with pytest.raises(ValueError, match="different length"):
        left > right

    with pytest.raises(ValueError, match="different length"):
        left >= right

    with pytest.raises(ValueError, match="different length"):
        left & right

    with pytest.raises(ValueError, match="different length"):
        left | right

    with pytest.raises(ValueError, match="different length"):
        left ^ right


def test_logical_ops_with_non_boolean() -> None:
    l_arrow = make_array([False, True, None, None])
    r_arrow = make_array([1, 2, 3, 4])

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    with pytest.raises(ValueError, match="logic"):
        left & right

    with pytest.raises(ValueError, match="logic"):
        left | right

    with pytest.raises(ValueError, match="logic"):
        left ^ right

    with pytest.raises(ValueError, match="logic"):
        right & left

    with pytest.raises(ValueError, match="logic"):
        right | left

    with pytest.raises(ValueError, match="logic"):
        right ^ left


def test_comparisons_dates() -> None:
    from datetime import date

    left = Series.from_pylist(
        [
            date(2023, 1, 1),
            date(2023, 1, 2),
            date(2023, 1, 3),
            None,
            date(2023, 1, 5),
            None,
        ]
    )
    right = Series.from_pylist(
        [
            date(2023, 1, 1),
            date(2023, 1, 3),
            date(2023, 1, 1),
            date(2023, 1, 5),
            None,
            None,
        ]
    )

    # eq, lt, gt, None, None, None

    lt = (left < right).to_pylist()
    assert lt == [False, True, False, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, None, None]


@pytest.mark.parametrize("l_dtype, r_dtype", itertools.product(arrow_binary_types, repeat=2))
def test_comparisons_binary(l_dtype, r_dtype) -> None:
    l_arrow = make_array([b"1", b"22", b"333", None, b"55555", None])
    r_arrow = make_array([b"1", b"333", b"1", b"55555", None, None])
    # eq, lt, gt, None, None, None

    left = Series.from_arrow(l_arrow.cast(l_dtype))
    right = Series.from_arrow(r_arrow.cast(r_dtype))
    lt = (left < right).to_pylist()
    assert lt == [False, True, False, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, None, None]


@pytest.mark.parametrize("l_dtype, r_dtype", itertools.product(arrow_binary_types, repeat=2))
def test_comparisons_binary_left_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([b"22"])
    r_arrow = make_array([b"1", b"22", b"333", None])
    # gt, eq, lt

    left = Series.from_arrow(l_arrow.cast(l_dtype))
    right = Series.from_arrow(r_arrow.cast(r_dtype))

    lt = (left < right).to_pylist()
    assert lt == [False, False, True, None]

    le = (left <= right).to_pylist()
    assert le == [False, True, True, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, True, False, None]

    gt = (left > right).to_pylist()
    assert gt == [True, False, False, None]


@pytest.mark.parametrize("l_dtype, r_dtype", itertools.product(arrow_binary_types, repeat=2))
def test_comparisons_binary_right_scalar(l_dtype, r_dtype) -> None:
    l_arrow = make_array([b"1", b"22", b"333", None, b"55555", None])
    r_arrow = make_array([b"22"])
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow.cast(l_dtype))
    right = Series.from_arrow(r_arrow.cast(r_dtype))
    lt = (left < right).to_pylist()
    assert lt == [True, False, False, None, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, False, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [False, True, True, None, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, True, None]


def test_comparisons_fixed_size_binary() -> None:
    l_arrow = make_array([b"11111", b"22222", b"33333", None, b"12345", None], type=pa.binary(5))
    r_arrow = make_array([b"11111", b"33333", b"11111", b"12345", None, None], type=pa.binary(5))
    # eq, lt, gt, None, None, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [False, True, False, None, None, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, None, None]

    eq = (left == right).to_pylist()
    assert eq == [True, False, False, None, None, None]

    neq = (left != right).to_pylist()
    assert neq == [False, True, True, None, None, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, False, True, None, None, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, None, None]


def test_comparisons_fixed_size_binary_left_scalar() -> None:
    l_arrow = make_array([b"222"], type=pa.binary(3))
    r_arrow = make_array([b"111", b"222", b"333", None], type=pa.binary(3))
    # gt, eq, lt

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)

    lt = (left < right).to_pylist()
    assert lt == [False, False, True, None]

    le = (left <= right).to_pylist()
    assert le == [False, True, True, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [True, True, False, None]

    gt = (left > right).to_pylist()
    assert gt == [True, False, False, None]


def test_comparisons_fixed_size_binary_right_scalar() -> None:
    l_arrow = make_array([b"111", b"222", b"333", None, b"555", None], type=pa.binary(3))
    r_arrow = make_array([b"222"], type=pa.binary(3))
    # lt, eq, gt, None, gt, None

    left = Series.from_arrow(l_arrow)
    right = Series.from_arrow(r_arrow)
    lt = (left < right).to_pylist()
    assert lt == [True, False, False, None, False, None]

    le = (left <= right).to_pylist()
    assert le == [True, True, False, None, False, None]

    eq = (left == right).to_pylist()
    assert eq == [False, True, False, None, False, None]

    neq = (left != right).to_pylist()
    assert neq == [True, False, True, None, True, None]

    ge = (left >= right).to_pylist()
    assert ge == [False, True, True, None, True, None]

    gt = (left > right).to_pylist()
    assert gt == [False, False, True, None, True, None]


class CustomZero:
    def __eq__(self, other):
        if isinstance(other, CustomZero):
            other = 0
        return 0 == other

    def __lt__(self, other):
        if isinstance(other, CustomZero):
            other = 0
        return 0 < other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)


@pytest.mark.parametrize(
    ["op", "reflected_op", "expected", "expected_self"],
    [
        (
            operator.eq,
            operator.eq,
            [False, True, False, None, None],
            [True, True, True, True, None],
        ),
        (
            operator.ne,
            operator.ne,
            [True, False, True, None, None],
            [False, False, False, False, None],
        ),
        (
            operator.lt,
            operator.gt,
            [False, False, True, None, None],
            [False, False, False, False, None],
        ),
        (
            operator.gt,
            operator.lt,
            [True, False, False, None, None],
            [False, False, False, False, None],
        ),
        (
            operator.le,
            operator.ge,
            [False, True, True, None, None],
            [True, True, True, True, None],
        ),
        (
            operator.ge,
            operator.le,
            [True, True, False, None, None],
            [True, True, True, True, None],
        ),
    ],
)
def test_comparisons_pyobjects(op, reflected_op, expected, expected_self) -> None:
    custom_zeros = Series.from_pylist([CustomZero(), CustomZero(), CustomZero(), CustomZero(), None])
    values = Series.from_pylist([-1, 0, 1, None, None])

    assert op(custom_zeros, values).datatype() == DataType.bool()
    assert op(custom_zeros, values).to_pylist() == expected
    assert op(custom_zeros, values).to_pylist() == reflected_op(values, custom_zeros).to_pylist()
    assert op(custom_zeros, custom_zeros).to_pylist() == expected_self


class CustomFalse:
    def __and__(self, other):
        if isinstance(other, CustomFalse):
            other = False
        return False & other

    def __or__(self, other):
        if isinstance(other, CustomFalse):
            other = False
        return False or other

    def __xor__(self, other):
        if isinstance(other, CustomFalse):
            other = False
        return False ^ other


@pytest.mark.parametrize(
    ["op", "expected", "expected_self"],
    [
        (operator.and_, [False, False, None, None], [False, False, False, None]),
        (operator.or_, [False, True, None, None], [False, False, False, None]),
        (operator.xor, [False, True, None, None], [False, False, False, None]),
    ],
)
def test_logicalops_pyobjects(op, expected, expected_self) -> None:
    custom_falses = Series.from_pylist([CustomFalse(), CustomFalse(), CustomFalse(), None])
    values = Series.from_pylist([False, True, None, None])

    # (Symmetry is not tested here since Python logicalops are not automatically symmetric.)
    assert op(custom_falses, values).datatype() == DataType.bool()
    assert op(custom_falses, values).to_pylist() == expected
    assert op(custom_falses, custom_falses).to_pylist() == expected_self


@pytest.mark.parametrize("tu1, tu2", itertools.product(["ns", "us", "ms"], repeat=2))
def test_compare_timestamps_no_tz(tu1, tu2):
    tz1 = Series.from_pylist([datetime(2022, 1, 1)])
    assert (tz1.cast(DataType.timestamp(tu1)) == tz1.cast(DataType.timestamp(tu2))).to_pylist() == [True]


def test_compare_timestamps_no_tz_date():
    tz1 = Series.from_pylist([datetime(2022, 1, 1)])
    Series.from_pylist([date(2022, 1, 1)])
    assert (tz1 == tz1).to_pylist() == [True]


def test_compare_timestamps_one_tz():
    tz1 = Series.from_pylist([datetime(2022, 1, 1)])
    tz2 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)])
    with pytest.raises(ValueError, match="Cannot perform comparison on types"):
        assert (tz1 == tz2).to_pylist() == [True]


def test_compare_timestamps_and_int():
    tz1 = Series.from_pylist([datetime(2022, 1, 1)])
    tz2 = Series.from_pylist([5])
    with pytest.raises(ValueError, match="Cannot perform comparison on types"):
        assert (tz1 == tz2).to_pylist() == [True]


def test_compare_timestamps_tz_date():
    tz1 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)])
    tz2 = Series.from_pylist([date(2022, 1, 1)])
    assert (tz1 == tz2).to_pylist() == [True]


def test_compare_lt_timestamps_tz_date():
    tz1 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)])
    tz2 = Series.from_pylist([date(2022, 1, 6)])
    assert (tz1 < tz2).to_pylist() == [True]


def test_compare_lt_timestamps_tz_date_same():
    tz1 = Series.from_pylist([datetime(2022, 1, 6, tzinfo=pytz.utc)])
    tz2 = Series.from_pylist([date(2022, 1, 6)])
    assert (tz1 < tz2).to_pylist() == [False]


@pytest.mark.parametrize("tu1, tu2", itertools.product(["ns", "us", "ms"], repeat=2))
def test_compare_timestamps_same_tz(tu1, tu2):
    tz1 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)]).cast(DataType.timestamp(tu1, "UTC"))
    tz2 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)]).cast(DataType.timestamp(tu2, "UTC"))
    assert (tz1 == tz2).to_pylist() == [True]


@pytest.mark.parametrize("tu1, tu2", itertools.product(["ns", "us", "ms"], repeat=2))
def test_compare_timestamps_diff_tz(tu1, tu2):
    utc = datetime(2022, 1, 1, tzinfo=pytz.utc)
    eastern = utc.astimezone(pytz.timezone("US/Eastern"))
    tz1 = Series.from_pylist([utc]).cast(DataType.timestamp(tu1, "UTC"))
    tz2 = Series.from_pylist([eastern]).cast(DataType.timestamp(tu1, "US/Eastern"))
    assert (tz1 == tz2).to_pylist() == [True]


@pytest.mark.parametrize("tu1, tu2", itertools.product(["ns", "us", "ms"], repeat=2))
def test_compare_lt_timestamps_same_tz(tu1, tu2):
    tz1 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)]).cast(DataType.timestamp(tu1, "UTC"))
    tz2 = Series.from_pylist([datetime(2022, 1, 1, tzinfo=pytz.utc)]).cast(DataType.timestamp(tu2, "UTC"))
    assert (tz1 < tz2).to_pylist() == [False]


@pytest.mark.parametrize("op", [operator.eq, operator.ne, operator.lt, operator.gt, operator.le, operator.ge])
def test_numeric_and_string_compare_raises_error(op):
    left = Series.from_pylist([1, 2, 3])
    right = Series.from_pylist(["1", "2", "3"])
    with pytest.raises(ValueError, match="Cannot perform comparison on types:"):
        op(left, right)

    with pytest.raises(ValueError, match="Cannot perform comparison on types:"):
        op(right, left)



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.daft import (
    LocalPhysicalPlan,
)
from daft.daft import (
    NativeExecutor as _NativeExecutor,
)
from daft.dataframe.display import MermaidOptions
from daft.recordbatch import MicroPartition

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.daft import PyDaftExecutionConfig
    from daft.logical.builder import LogicalPlanBuilder
    from daft.runners.partitioning import (
        LocalMaterializedResult,
        MaterializedResult,
        PartitionT,
    )


class NativeExecutor:
    def __init__(self) -> None:
        self._executor = _NativeExecutor()

    def run(
        self,
        local_physical_plan: LocalPhysicalPlan,
        psets: dict[str, list[MaterializedResult[PartitionT]]],
        daft_execution_config: PyDaftExecutionConfig,
        results_buffer_size: int | None,
    ) -> Iterator[LocalMaterializedResult]:
        from daft.runners.partitioning import LocalMaterializedResult

        psets_mp = {
            part_id: [part.micropartition()._micropartition for part in parts] for part_id, parts in psets.items()
        }
        return (
            LocalMaterializedResult(MicroPartition._from_pymicropartition(part))
            for part in self._executor.run(
                local_physical_plan,
                psets_mp,
                daft_execution_config,
                results_buffer_size,
            )
        )

    def pretty_print(
        self,
        builder: LogicalPlanBuilder,
        daft_execution_config: PyDaftExecutionConfig,
        simple: bool = False,
        format: str = "ascii",
    ) -> str:
        """Pretty prints the current underlying logical plan."""
        if format == "ascii":
            return self._executor.repr_ascii(builder._builder, daft_execution_config, simple)
        elif format == "mermaid":
            return self._executor.repr_mermaid(builder._builder, daft_execution_config, MermaidOptions(simple))
        else:
            raise ValueError(f"Unknown format: {format}")



from __future__ import annotations

import pytest

from daft.expressions import col
from tests.conftest import assert_df_equals


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_distinct_all_columns(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Sums across groups."""
    daft_df = daft_df.repartition(repartition_nparts).select(*[col(k) for k in keys]).distinct()

    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.groupby(keys).sum("Unique Key").reset_index().drop("Unique Key", axis=1)
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.lazy_import import LazyImport

if TYPE_CHECKING:
    import fsspec
    import numpy as np
    import pandas as pd
    import PIL.Image as pil_image
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.csv as pacsv
    import pyarrow.dataset as pads
    import pyarrow.flight as flight
    import pyarrow.fs as pafs
    import pyarrow.json as pajson
    import pyarrow.parquet as pq
else:
    fsspec = LazyImport("fsspec")
    np = LazyImport("numpy")
    pd = LazyImport("pandas")
    pil_image = LazyImport("PIL.Image")
    pa = LazyImport("pyarrow")
    pacsv = LazyImport("pyarrow.csv")
    pads = LazyImport("pyarrow.dataset")
    pafs = LazyImport("pyarrow.fs")
    pajson = LazyImport("pyarrow.json")
    pc = LazyImport("pyarrow.compute")
    pq = LazyImport("pyarrow.parquet")
    flight = LazyImport("pyarrow.flight")

unity_catalog = LazyImport("daft.unity_catalog")

__all__ = [
    "flight",
    "fsspec",
    "np",
    "pa",
    "pacsv",
    "pads",
    "pafs",
    "pajson",
    "pc",
    "pd",
    "pil_image",
    "pq",
    "unity_catalog",
]



from __future__ import annotations

from fastapi import Response


def get_response(url: str, status_code: int, code_str: str):
    return Response(
        status_code=status_code,
        content=f"""<?xml version="1.0" encoding="UTF-8"?>
<Error>
<Code>{code_str}</Code>
<Message>{code_str}</Message>
<Resource>{url}</Resource>
<RequestId>4442587FB7D0A2F9</RequestId>
</Error>""",
        headers={
            "Content-Type": "application/xml",
        },
    )



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, ScanOperatorHandle, StorageConfig
from daft.dataframe import DataFrame
from daft.dependencies import unity_catalog
from daft.io.catalog import DataCatalogTable
from daft.logical.builder import LogicalPlanBuilder

if TYPE_CHECKING:
    from datetime import datetime

    from daft.unity_catalog import UnityCatalogTable


@PublicAPI
def read_deltalake(
    table: Union[str, DataCatalogTable, "UnityCatalogTable"],
    version: Optional[Union[int, str, "datetime"]] = None,
    io_config: Optional[IOConfig] = None,
    _multithreaded_io: Optional[bool] = None,
) -> DataFrame:
    """Create a DataFrame from a Delta Lake table.

    Args:
        table: Either a URI for the Delta Lake table or a :class:`~daft.io.catalog.DataCatalogTable` instance
            referencing a table in a data catalog, such as AWS Glue Data Catalog or Databricks Unity Catalog.
        version (optional): If int is passed, read the table with specified version number. Otherwise if string or datetime,
            read the timestamp version of the table. Strings must be RFC 3339 and ISO 8601 date and time format.
            Datetimes are assumed to be UTC timezone unless specified. By default, read the latest version of the table.
        io_config (optional): A custom :class:`~daft.daft.IOConfig` to use when accessing Delta Lake object storage data. Defaults to None.
        _multithreaded_io (optional): Whether to use multithreading for IO threads. Setting this to False can be helpful in reducing
            the amount of system resources (number of connections and thread contention) when running in the Ray runner.
            Defaults to None, which will let Daft decide based on the runner it is currently using.

    Returns:
        DataFrame: A DataFrame with the schema converted from the specified Delta Lake table.

    Note:
        This function requires the use of [deltalake](https://delta-io.github.io/delta-rs/), a Python library for interacting with Delta Lake.

    Examples:
        >>> df = daft.read_deltalake("some-table-uri")
        >>>
        >>> # Filters on this dataframe can now be pushed into
        >>> # the read operation from Delta Lake.
        >>> df = df.where(df["foo"] > 5)
        >>> df.show()
    """
    from daft.io.delta_lake.delta_lake_scan import DeltaLakeScanOperator

    # If running on Ray, we want to limit the amount of concurrency and requests being made.
    # This is because each Ray worker process receives its own pool of thread workers and connections
    multithreaded_io = (
        (context.get_context().get_or_create_runner().name != "ray") if _multithreaded_io is None else _multithreaded_io
    )

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_config = StorageConfig(multithreaded_io, io_config)

    if isinstance(table, str):
        table_uri = table
    elif isinstance(table, DataCatalogTable):
        table_uri = table.table_uri(io_config)
    elif unity_catalog.module_available() and isinstance(table, unity_catalog.UnityCatalogTable):
        table_uri = table.table_uri

        # Override the storage_config with the one provided by Unity catalog
        recordbatch_io_config = table.io_config
        if recordbatch_io_config is not None:
            storage_config = StorageConfig(multithreaded_io, recordbatch_io_config)
    else:
        raise ValueError(
            f"table argument must be a table URI string, DataCatalogTable or UnityCatalogTable instance, but got: {type(table)}, {table}"
        )
    delta_lake_operator = DeltaLakeScanOperator(table_uri, storage_config=storage_config, version=version)

    handle = ScanOperatorHandle.from_python_scan_operator(delta_lake_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)


def large_dtypes_kwargs(large_dtypes: bool) -> dict[str, Any]:
    import deltalake
    from packaging.version import parse

    if parse(deltalake.__version__) < parse("0.19.0"):
        return {"large_dtypes": large_dtypes}
    else:
        from deltalake.schema import ArrowSchemaConversionMode

        schema_conversion_mode = ArrowSchemaConversionMode.LARGE if large_dtypes else ArrowSchemaConversionMode.NORMAL
        return {"schema_conversion_mode": schema_conversion_mode}



from __future__ import annotations

import pathlib
import sqlite3

import pandas as pd
import pytest

from benchmarking.tpch import data_generation
from tests.assets import TPCH_DBGEN_DIR, TPCH_QUERIES
from tests.conftest import assert_df_equals

# Hardcode scale factor to 200M for local testing
SCALE_FACTOR = 0.2


@pytest.fixture(scope="session", params=[1, 2])
def gen_tpch(request):
    # Parametrize the number of parts for each file so that we run tests on single-partition files and multi-partition files
    num_parts = request.param

    csv_files_location = data_generation.gen_csv_files(TPCH_DBGEN_DIR, num_parts, SCALE_FACTOR)

    sqlite_path = data_generation.gen_sqlite_db(
        csv_filepath=csv_files_location,
        num_parts=num_parts,
    )

    return csv_files_location, sqlite_path


@pytest.fixture(scope="module")
def check_answer(gen_tpch):
    _, sqlite_db_file_path = gen_tpch

    def _check_answer(daft_pd_df: pd.DataFrame, tpch_question: int, tmp_path: str):
        query = pathlib.Path(f"{TPCH_QUERIES}/{tpch_question}.sql").read_text()
        conn = sqlite3.connect(sqlite_db_file_path, detect_types=sqlite3.PARSE_DECLTYPES)
        cursor = conn.cursor()
        queries = query.split(";")
        for q in queries:
            if not q.isspace():
                res = cursor.execute(q)
        sqlite_results = res.fetchall()
        sqlite_pd_results = pd.DataFrame.from_records(sqlite_results, columns=daft_pd_df.columns)
        assert_df_equals(
            daft_pd_df,
            sqlite_pd_results,
            assert_ordering=True,
            # We lose fine-grained dtype information from the sqlite3 API
            check_dtype=False,
        )

    return _check_answer



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.series import Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES


@pytest.mark.parametrize("inner_dtype", ARROW_FLOAT_TYPES + ARROW_INT_TYPES)
@pytest.mark.parametrize("fixed", [False, True])
@pytest.mark.parametrize("desc", [False, True])
def test_list_list_sort(inner_dtype, fixed, desc):
    dtype = pa.list_(inner_dtype, list_size=2 if fixed else -1)
    data_list = [
        [20, 10],
        [None, 50],
        [45, 43],
        None,
        [43, 45],
        [50, 50],
        [80, None],
        [50, 52],
        None,
    ]
    data = pa.array(data_list, type=dtype)

    s = Series.from_arrow(data)
    res = s.list.sort(desc)
    assert s.datatype() == res.datatype()

    sorted_data_list = [
        (sorted(x, key=lambda v: (v is None, v), reverse=desc) if x is not None else x) for x in data_list
    ]
    assert res.to_pylist() == sorted_data_list


@pytest.mark.parametrize("inner_dtype", ARROW_STRING_TYPES)
@pytest.mark.parametrize("fixed", [False, True])
@pytest.mark.parametrize("desc", [False, True])
def test_list_list_sort_strings(inner_dtype, fixed, desc):
    dtype = pa.list_(inner_dtype, list_size=2 if fixed else -1)
    data_list = [
        ["hello", "world"],
        ["aaaa", "aaa"],
        ["aa", "aaaa"],
        [None, "hi"],
        None,
        ["", "blahblah"],
        ["None", None],
        None,
    ]
    data = pa.array(data_list, type=dtype)

    s = Series.from_arrow(data)
    res = s.list.sort(desc)
    assert s.datatype() == res.datatype()

    sorted_data_list = [
        (sorted(x, key=lambda v: (v is None, v), reverse=desc) if x is not None else None) for x in data_list
    ]
    assert res.to_pylist() == sorted_data_list


@pytest.mark.parametrize("fixed", [False, True])
def test_list_list_sort_multi_desc(fixed):
    dtype = pa.list_(pa.int64(), list_size=2 if fixed else -1)
    data_list = [
        [10, 20],
        [10, 20],
        [40, 30],
        [40, 30],
    ]
    data = pa.array(data_list, type=dtype)

    desc = Series.from_pylist([False, True, False, True])

    s = Series.from_arrow(data)
    res = s.list.sort(desc)
    assert s.datatype() == res.datatype()

    expected = [
        [10, 20],
        [20, 10],
        [30, 40],
        [40, 30],
    ]
    assert res.to_pylist() == expected



from __future__ import annotations

from typing import Any

from daft.daft import PyWindowBoundary as _PyWindowBoundary
from daft.daft import WindowFrame as _WindowFrame
from daft.daft import WindowSpec as _WindowSpec
from daft.expressions import Expression
from daft.utils import ManyColumnsInputType, column_inputs_to_expressions


class Window:
    """Describes how to partition data and in what order to apply the window function.

    This class provides a way to specify window definitions for window functions.
    Window functions operate on a group of rows (called a window frame) and return
    a result for each row based on the values in its window frame.

    Examples:
        >>> from daft import Window, col
        >>>
        >>> # Basic window aggregation with a single partition column:
        >>> window_spec = Window().partition_by("category")
        >>> df = df.select(
        ...     col("value").sum().over(window_spec).alias("category_total"),
        ...     col("value").mean().over(window_spec).alias("category_avg"),
        ... )
        >>>
        >>> # Partitioning by multiple columns:
        >>> window_spec = Window().partition_by(["department", "category"])
        >>> df = df.select(col("sales").sum().over(window_spec).alias("dept_category_total"))
        >>>
        >>> # Using window aggregations in expressions:
        >>> window_spec = Window().partition_by("category")
        >>> df = df.select((col("value") / col("value").sum().over(window_spec)).alias("pct_of_category"))
    """

    # Class-level constants for frame boundaries
    unbounded_preceding = _PyWindowBoundary.unbounded_preceding()
    unbounded_following = _PyWindowBoundary.unbounded_following()
    current_row = _PyWindowBoundary.offset(0)

    def __init__(self) -> None:
        self._spec = _WindowSpec.new()

    def partition_by(self, *cols: ManyColumnsInputType) -> Window:
        """Partitions the dataset by one or more columns or expressions.

        Args:
            *cols: Columns or expressions on which to partition data.
                   Can be column names as strings, Expression objects, or iterables of these.

        Returns:
            Window: A window specification with the given partitioning.

        Raises:
            ValueError: If no partition columns are specified.

        Examples:
            >>> from daft import Window, col
            >>> # Partition by a single column 'category'
            >>> window_spec = Window().partition_by("category")
            >>> # Partition by multiple columns 'department' and 'region'
            >>> window_spec_multi = Window().partition_by("department", "region")
        """
        if not cols:
            raise ValueError("At least one partition column must be specified")

        expressions = []
        for c in cols:
            expressions.extend(column_inputs_to_expressions(c))

        if not expressions:
            raise ValueError("At least one partition column must be specified")

        window = Window()
        window._spec = self._spec.with_partition_by([expr._expr for expr in expressions])
        return window

    def order_by(
        self, *cols: ManyColumnsInputType, desc: bool | list[bool] = False, nulls_first: bool | list[bool] | None = None
    ) -> Window:
        """Orders rows within each partition by specified columns or expressions.

        Args:
            *cols: Columns or expressions to determine ordering within the partition.
                   Can be column names as strings, Expression objects, or iterables of these.
            desc: Sort descending (True) or ascending (False). Can be a single boolean value applied to all columns,
                 or a list of boolean values corresponding to each column. Default is False (ascending).
            nulls_first: Whether to position NULL values at the beginning (True) or end (False) of the partition.
                Can be a single boolean value applied to all columns, or a list of boolean values corresponding to each column.
                Default is None, which means NULL values are positioned at the end for ascending order (default) and at the beginning for descending order.

        Returns:
            Window: A window specification with the given ordering.

        Examples:
            >>> from daft import Window, col
            >>> # Order by 'date' ascending (default)
            >>> window_spec = Window().partition_by("category").order_by("date")
            >>> # Order by 'sales' descending
            >>> window_spec_desc = Window().partition_by("category").order_by("sales", desc=True)
            >>> # Order by 'date' ascending and 'sales' descending
            >>> window_spec_multi = Window().partition_by("category").order_by("date", "sales", desc=[False, True])
        """
        expressions = []
        for c in cols:
            expressions.extend(column_inputs_to_expressions(c))

        if isinstance(desc, bool):
            desc_flags = [desc] * len(expressions)
        else:
            if len(desc) != len(expressions):
                raise ValueError("Length of descending flags must match number of order by columns")
            desc_flags = desc

        if nulls_first is None:
            nulls_first_flags = desc_flags
        elif isinstance(nulls_first, bool):
            nulls_first_flags = [nulls_first] * len(expressions)
        else:
            if len(nulls_first) != len(expressions):
                raise ValueError("Length of nulls first flags must match number of order by columns")
            nulls_first_flags = nulls_first

        window = Window()
        window._spec = self._spec.with_order_by([expr._expr for expr in expressions], desc_flags, nulls_first_flags)
        return window

    def rows_between(
        self,
        start: int | _PyWindowBoundary,
        end: int | _PyWindowBoundary,
        min_periods: int = 1,
    ) -> Window:
        """Restricts each window to a row-based frame between start and end boundaries.

        This defines a sliding window based on row offsets relative to the current row.

        Args:
            start: Boundary definitions for the start of the window. Can be:

                *   ``Window.unbounded_preceding``: Include all rows before the current row.
                *   ``Window.current_row``: Start at the current row.
                *   Integer value (e.g., ``-3``): A negative integer indicates the number of rows *preceding* the current row.
                *   Integer value (e.g., ``1``): A positive integer indicates the number of rows *following* the current row.
            end: Boundary definitions for the end of the window. Can be:

                *   ``Window.unbounded_following``: Include all rows after the current row.
                *   ``Window.current_row``: End at the current row.
                *   Integer value (e.g., ``1``): A positive integer indicates the number of rows *following* the current row.
                *   Integer value (e.g., ``-1``): A negative integer indicates the number of rows *preceding* the current row.
            min_periods: Minimum number of rows required in the window frame to compute a result (default = 1).
                If fewer rows exist in the frame, the function returns NULL.

        Returns:
            Window: A window specification with the given row-based frame bounds.

        Examples:
            >>> from daft import Window
            >>> # Frame includes the current row and the 2 preceding rows
            >>> window_spec = Window().partition_by("cat").order_by("val").rows_between(-2, Window.current_row)
            >>> # Frame includes all rows from the beginning of the partition up to the current row
            >>> cum_window = (
            ...     Window()
            ...     .partition_by("cat")
            ...     .order_by("val")
            ...     .rows_between(Window.unbounded_preceding, Window.current_row)
            ... )
            >>> # Frame includes the preceding row, current row, and following row
            >>> sliding_window = Window().partition_by("cat").order_by("val").rows_between(-1, 1)
            >>> # Frame includes the current row and the 3 following rows, requiring at least 2 rows
            >>> lookahead_window = (
            ...     Window().partition_by("cat").order_by("val").rows_between(Window.current_row, 3, min_periods=2)
            ... )
        """
        if isinstance(start, int):
            start = _PyWindowBoundary.offset(start)
        if isinstance(end, int):
            end = _PyWindowBoundary.offset(end)

        frame = _WindowFrame(
            start=start,
            end=end,
        )

        new_window = Window()
        new_window._spec = self._spec.with_frame(frame).with_min_periods(min_periods)
        return new_window

    def range_between(
        self,
        start: Any,
        end: Any,
        min_periods: int = 1,
    ) -> Window:
        """Restricts each window to a range-based frame between start and end boundaries.

        This defines a window frame based on a range of values relative to the current row's
        value in the ordering column. Requires exactly one `order_by` column, which must be
        numeric or temporal type.

        Args:
            start: Boundary definition for the start of the window's range. Can be:

                *   ``Window.unbounded_preceding``: Include all rows with order value <= current row's order value + start offset.
                *   ``Window.current_row``: Start range at the current row's order value.
                *   Offset value (e.g., ``-10``, ``datetime.timedelta(days=-1)``): The start of the range is defined as
                    `current_row_order_value + start`. The type of the offset must match the order-by column type.
                    Negative values indicate a lower bound *less than* the current row's value. Positive values indicate a lower bound
                    *more than* the current row's value.
            end: Boundary definition for the end of the window's range. Syntax is similar to `start`.
                Negative values indicate an upper bound *less than* the current row's value. Positive values indicate an upper bound
                *more than* the current row's value.
            min_periods: Minimum number of rows required in the window frame to compute a result (default = 1).
                If fewer rows exist in the frame, the function returns NULL.

        Returns:
            Window: A window specification with the given range-based frame bounds.

        Raises:
            NotImplementedError: This feature is not yet implemented.

        Examples:
            >>> from daft import Window, col
            >>> import datetime
            >>> # Assume df has columns 'sensor_id', 'timestamp', 'reading'
            >>> # Frame includes rows within 10 units *before* the current row's reading
            >>> val_window = Window().partition_by("sensor_id").order_by("reading").range_between(-10, Window.current_row)
            >>> # Frame includes rows from 1 day before to 1 day after the current row's timestamp
            >>> time_window = (
            ...     Window()
            ...     .partition_by("sensor_id")
            ...     .order_by("timestamp")
            ...     .range_between(datetime.timedelta(days=-1), datetime.timedelta(days=1))
            ... )
        """
        if isinstance(start, _PyWindowBoundary):
            start_boundary = start
        else:
            start_expr = Expression._to_expression(start)
            start_boundary = _PyWindowBoundary.range_offset(start_expr._expr)

        if isinstance(end, _PyWindowBoundary):
            end_boundary = end
        else:
            end_expr = Expression._to_expression(end)
            end_boundary = _PyWindowBoundary.range_offset(end_expr._expr)

        frame = _WindowFrame(
            start=start_boundary,
            end=end_boundary,
        )

        new_window = Window()
        new_window._spec = self._spec.with_frame(frame).with_min_periods(min_periods)
        return new_window



from __future__ import annotations

import copy

import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.expressions import ExpressionsProjection, col
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition

DATA = {
    "int": ([1, 2, None], DataType.int64()),
    "float": ([1.0, 2.0, None], DataType.float64()),
    "string": (["a", "b", None], DataType.string()),
    "bool": ([True, True, None], DataType.bool()),
}

TABLE = MicroPartition.from_pydict({k: data for k, (data, _) in DATA.items()})
EXPECTED_TYPES = {k: t for k, (_, t) in DATA.items()}

from tests.utils import ANSI_ESCAPE, TD_STYLE, TH_STYLE


def test_schema_len():
    schema = TABLE.schema()
    assert len(schema) == len(DATA)


def test_schema_column_names():
    schema = TABLE.schema()
    assert schema.column_names() == list(DATA.keys())


def test_schema_field_types():
    schema = TABLE.schema()
    for key in EXPECTED_TYPES:
        assert schema[key].name == key
        assert schema[key].dtype == EXPECTED_TYPES[key]


def test_schema_iter():
    schema = TABLE.schema()
    for expected_name, field in zip(EXPECTED_TYPES, schema):
        assert field.name == expected_name
        assert field.dtype == EXPECTED_TYPES[expected_name]


def test_schema_eq():
    t1, t2 = (
        MicroPartition.from_pydict({k: data for k, (data, _) in DATA.items()}),
        MicroPartition.from_pydict({k: data for k, (data, _) in DATA.items()}),
    )
    s1, s2 = t1.schema(), t2.schema()
    assert s1 == s2

    t_empty = MicroPartition.empty()
    assert s1 != t_empty.schema()


def test_schema_to_name_set():
    schema = TABLE.schema()
    assert schema.to_name_set() == set(DATA.keys())


def test_truncated_repr():
    schema = TABLE.schema()
    out_repr = schema._truncated_table_string()
    without_escape = ANSI_ESCAPE.sub("", out_repr)
    assert (
        without_escape.replace("\r", "")
        == """╭───────┬─────────┬────────┬─────────╮
│ int   ┆ float   ┆ string ┆ bool    │
│ ---   ┆ ---     ┆ ---    ┆ ---     │
│ Int64 ┆ Float64 ┆ Utf8   ┆ Boolean │
╰───────┴─────────┴────────┴─────────╯
"""
    )


def test_truncated_repr_html():
    schema = TABLE.schema()
    out_repr = schema._truncated_table_html()
    assert (
        out_repr
        == f"""<table class="dataframe">
<thead><tr><th {TH_STYLE}>int<br />Int64</th><th {TH_STYLE}>float<br />Float64</th><th {TH_STYLE}>string<br />Utf8</th><th {TH_STYLE}>bool<br />Boolean</th></tr></thead>
</table>"""
    )


def test_repr():
    schema = TABLE.schema()
    out_repr = repr(schema)
    without_escape = ANSI_ESCAPE.sub("", out_repr)
    assert (
        without_escape.replace("\r", "")
        == """╭─────────────┬─────────╮
│ column_name ┆ type    │
╞═════════════╪═════════╡
│ int         ┆ Int64   │
├╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
│ float       ┆ Float64 │
├╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
│ string      ┆ Utf8    │
├╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
│ bool        ┆ Boolean │
╰─────────────┴─────────╯
"""
    )


def test_repr_html():
    schema = TABLE.schema()
    out_repr = schema._repr_html_()
    assert (
        out_repr
        == f"""<table class="dataframe">
<thead><tr><th {TH_STYLE}>column_name</th><th {TH_STYLE}>type</th></tr></thead>
<tbody>
<tr><td {TD_STYLE}>int</td><td {TD_STYLE}>Int64</td></tr>
<tr><td {TD_STYLE}>float</td><td {TD_STYLE}>Float64</td></tr>
<tr><td {TD_STYLE}>string</td><td {TD_STYLE}>Utf8</td></tr>
<tr><td {TD_STYLE}>bool</td><td {TD_STYLE}>Boolean</td></tr>
</tbody>
</table>"""
    )


def test_to_col_expr():
    schema = TABLE.schema()
    schema_col_exprs = ExpressionsProjection.from_schema(schema)
    expected_col_exprs = [col(n) for n in schema.column_names()]

    assert len(schema_col_exprs) == len(expected_col_exprs)
    for sce, ece in zip(schema_col_exprs, expected_col_exprs):
        assert sce.name() == ece.name()


def test_union():
    schema = TABLE.schema()
    with pytest.raises(ValueError):
        schema.union(schema)

    new_data = {f"{k}_": d for k, (d, _) in DATA.items()}
    new_table = MicroPartition.from_pydict(new_data)
    unioned_schema = schema.union(new_table.schema())

    assert unioned_schema.column_names() == list(DATA.keys()) + list(new_data.keys())
    assert list(unioned_schema) == list(schema) + list(new_table.schema())


def test_from_field_name_and_types():
    schema = Schema._from_field_name_and_types([("foo", DataType.int16())])
    assert schema["foo"].name == "foo"
    assert schema["foo"].dtype == DataType.int16()


def test_from_empty_field_name_and_types():
    schema = Schema._from_field_name_and_types([])
    assert len(schema) == 0


def test_field_pickling():
    schema = Schema._from_field_name_and_types([("foo", DataType.int16())])
    f1 = schema["foo"]
    f1_copy = copy.deepcopy(f1)
    assert f1_copy.dtype == f1.dtype
    assert f1_copy.name == f1.name
    assert f1_copy == f1


def test_schema_pickling():
    t1, t2 = (
        MicroPartition.from_pydict({k: data for k, (data, _) in DATA.items()}),
        MicroPartition.from_pydict({k: data for k, (data, _) in DATA.items()}),
    )

    s1, s2 = t1.schema(), t2.schema()

    s1 = copy.deepcopy(s1)

    assert s1 == s2

    t_empty = MicroPartition.empty()
    assert s1 != t_empty.schema()
    t_empty_schema_copy = copy.deepcopy(t_empty.schema())
    assert t_empty.schema() == t_empty_schema_copy


def test_schema_pyarrow_roundtrip():
    pa_schema = pa.schema(
        {
            "int": pa.int64(),
            "str": pa.string(),
            "list": pa.list_(pa.int64()),
            "map": pa.map_(pa.string(), pa.int64()),
        }
    )

    expected_daft_schema = Schema._from_field_name_and_types(
        [
            ("int", DataType.int64()),
            ("str", DataType.string()),
            ("list", DataType.list(DataType.int64())),
            ("map", DataType.map(DataType.string(), DataType.int64())),
        ]
    )

    assert Schema.from_pyarrow_schema(pa_schema) == expected_daft_schema

    roundtrip_pa_schema = pa.schema(
        {
            "int": pa.int64(),
            "str": pa.large_string(),
            "list": pa.large_list(pa.int64()),
            "map": pa.map_(pa.large_string(), pa.int64()),
        }
    )

    assert Schema.from_pyarrow_schema(pa_schema).to_pyarrow_schema() == roundtrip_pa_schema


def test_schema_pyarrow_roundtrip_with_metadata():
    metadata = {
        "lance-encoding:compression": "zstd",
        "lance-encoding:compression-level": "3",
        "lance-encoding:structural-encoding": "miniblock",
        "lance-encoding:packed": "true",
    }
    pa_schema = pa.schema(
        [
            pa.field("compressible_strings", pa.string(), metadata=metadata),
            pa.field("uncompressed_stuff", pa.int64()),
        ]
    )

    # Daft seems to default to large_string on conversion, so we create an expected schema with large_string
    expected_roundtrip_pa_schema = pa.schema(
        [
            pa.field("compressible_strings", pa.large_string(), metadata=metadata),
            pa.field("uncompressed_stuff", pa.int64()),
        ]
    )

    daft_schema = Schema.from_pyarrow_schema(pa_schema)
    roundtrip_pa_schema = daft_schema.to_pyarrow_schema()

    assert roundtrip_pa_schema == expected_roundtrip_pa_schema



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from enum import Enum
from typing import Any

from daft.dependencies import pa, pafs, pq
from daft.filesystem import join_path


class State(Enum):
    REQUESTED = 0
    INFLIGHT = 1
    COMPLETED = 2


@dataclass
class Instant:
    state: State
    action: str
    timestamp: str

    @property
    def file_name(self) -> str:
        state = "" if self.state == State.COMPLETED else f".{self.state.name.lower()}"
        return f"{self.timestamp}.{self.action}{state}"

    def __lt__(self, other: Instant) -> bool:
        return [self.timestamp, self.state] < [other.timestamp, other.state]


@dataclass(init=False)
class Timeline:
    base_path: str
    fs: pafs.FileSystem
    completed_commit_instants: list[Instant]

    def __init__(self, base_path: str, fs: pafs.FileSystem) -> None:
        self.base_path = base_path
        self.fs = fs
        self._load_completed_commit_instants()

    @property
    def has_completed_commit(self) -> bool:
        return len(self.completed_commit_instants) > 0

    def _load_completed_commit_instants(self) -> None:
        timeline_path = join_path(self.fs, self.base_path, ".hoodie")
        write_action_exts = {".commit"}
        commit_instants = []
        for file_info in self.fs.get_file_info(pafs.FileSelector(timeline_path)):
            ext = os.path.splitext(file_info.base_name)[1]
            if ext in write_action_exts:
                timestamp = file_info.base_name[: -len(ext)]
                instant = Instant(state=State.COMPLETED, action=ext[1:], timestamp=timestamp)
                commit_instants.append(instant)
        self.completed_commit_instants = sorted(commit_instants)

    def get_latest_commit_metadata(self) -> dict[Any, Any]:
        if not self.has_completed_commit:
            return {}
        latest_instant_file_path = join_path(
            self.fs, self.base_path, ".hoodie", self.completed_commit_instants[-1].file_name
        )
        with self.fs.open_input_file(latest_instant_file_path) as f:
            return json.load(f)

    def get_latest_commit_schema(self) -> pa.Schema:
        latest_commit_metadata = self.get_latest_commit_metadata()
        if not latest_commit_metadata.get("partitionToWriteStats"):
            return pa.schema([])

        _, write_stats = next(iter(latest_commit_metadata["partitionToWriteStats"].items()))
        base_file_path = join_path(self.fs, self.base_path, write_stats[0]["path"])
        with self.fs.open_input_file(base_file_path) as f:
            return pq.read_schema(f)



from __future__ import annotations

import pytest

import daft
from daft.context import get_context
from tests.conftest import get_tests_daft_runner_name


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
def test_active_plan_clean_up_df_show():
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        path = "tests/assets/parquet-data/mvp.parquet"
        df = daft.read_parquet([path, path])
        df.show()
        runner = get_context().get_or_create_runner()
        assert len(runner.active_plans()) == 0


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
def test_active_plan_single_iter_partitions():
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        path = "tests/assets/parquet-data/mvp.parquet"
        df = daft.read_parquet([path, path])
        iter = df.iter_partitions()
        next(iter)
        runner = get_context().get_or_create_runner()
        assert len(runner.active_plans()) == 1
        del iter
        assert len(runner.active_plans()) == 0


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
def test_active_plan_multiple_iter_partitions():
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        path = "tests/assets/parquet-data/mvp.parquet"
        df = daft.read_parquet([path, path])
        iter = df.iter_partitions()
        next(iter)
        runner = get_context().get_or_create_runner()
        assert len(runner.active_plans()) == 1

        df2 = daft.read_parquet([path, path])
        iter2 = df2.iter_partitions()
        next(iter2)
        assert len(runner.active_plans()) == 2

        del iter
        assert len(runner.active_plans()) == 1

        del iter2
        assert len(runner.active_plans()) == 0


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
def test_active_plan_with_show_and_write_parquet(tmpdir):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_parquet("tests/assets/parquet-data/mvp.parquet")
        df = df.into_partitions(8)
        df = df.join(df, on="a")
        df.show()
        runner = get_context().get_or_create_runner()
        assert len(runner.active_plans()) == 0
        df.write_parquet(tmpdir.dirname)
        assert len(runner.active_plans()) == 0



from __future__ import annotations

from daft.daft import PyPartitionField, PyPartitionTransform
from daft.schema import Field


class PartitionField:
    """Partitioning Field of a Scan Source such as Hive or Iceberg."""

    _partition_field: PyPartitionField

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a PartitionField via __init__ ")

    def __repr__(self) -> str:
        return self._partition_field.__repr__()

    @classmethod
    def create(
        cls,
        field: Field,
        source_field: Field | None = None,
        transform: PartitionTransform | None = None,
    ) -> PartitionField:
        pf = cls.__new__(cls)
        pf._partition_field = PyPartitionField(
            field._field,
            source_field._field if source_field is not None else None,
            transform._partition_transform if transform is not None else None,
        )
        return pf

    @property
    def field(self) -> Field:
        return Field._from_pyfield(self._partition_field.field)

    @property
    def source_field(self) -> Field | None:
        if source_field := self._partition_field.source_field:
            return Field._from_pyfield(source_field)
        else:
            return None

    @property
    def transform(self) -> PartitionTransform | None:
        if transform := self._partition_field.transform:
            return PartitionTransform._from_py_partition_transform(transform)
        else:
            return None


class PartitionTransform:
    """Partitioning Transform from a Data Catalog source field to a Partitioning Columns."""

    _partition_transform: PyPartitionTransform

    def __init__(self) -> None:
        raise ValueError(
            "We do not support creating a PartitionTransform directly, please use one of the factory methods."
        )

    def __repr__(self) -> str:
        return self._partition_transform.__repr__()

    def __eq__(self, value: object) -> bool:
        if isinstance(value, PartitionTransform):
            return self._partition_transform == value._partition_transform
        return False

    @classmethod
    def _from_py_partition_transform(cls, py_partition_transform: PyPartitionTransform) -> PartitionTransform:
        pt = cls.__new__(cls)
        pt._partition_transform = py_partition_transform
        return pt

    @classmethod
    def identity(cls) -> PartitionTransform:
        """Creates an identity partition transform that returns the input value unchanged.

        Returns:
            PartitionTransform: A new identity partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.identity()
            >>> assert transform.is_identity()
        """
        pt = PyPartitionTransform.identity()
        return cls._from_py_partition_transform(pt)

    @classmethod
    def year(cls) -> PartitionTransform:
        """Creates a partition transform that extracts the year from a timestamp.

        Returns:
            PartitionTransform: A new year partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.year()
            >>> assert transform.is_year()
        """
        pt = PyPartitionTransform.year()
        return cls._from_py_partition_transform(pt)

    @classmethod
    def month(cls) -> PartitionTransform:
        """Creates a partition transform that extracts the month from a timestamp.

        Returns:
            PartitionTransform: A new month partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.month()
            >>> assert transform.is_month()
        """
        pt = PyPartitionTransform.month()
        return cls._from_py_partition_transform(pt)

    @classmethod
    def day(cls) -> PartitionTransform:
        """Creates a partition transform that extracts the day from a timestamp.

        Returns:
            PartitionTransform: A new day partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.day()
            >>> assert transform.is_day()
        """
        pt = PyPartitionTransform.day()
        return cls._from_py_partition_transform(pt)

    @classmethod
    def hour(cls) -> PartitionTransform:
        """Creates a partition transform that extracts the hour from a timestamp.

        Returns:
            PartitionTransform: A new hour partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.hour()
            >>> assert transform.is_hour()
        """
        pt = PyPartitionTransform.hour()
        return cls._from_py_partition_transform(pt)

    @classmethod
    def iceberg_bucket(cls, n: int) -> PartitionTransform:
        """Creates an Iceberg bucket partition transform that hashes values into n buckets.

        Args:
            n: Number of buckets to hash values into.

        Returns:
            PartitionTransform: A new Iceberg bucket partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_bucket(10)
            >>> assert transform.is_iceberg_bucket()
            >>> assert transform.num_buckets == 10
        """
        pt = PyPartitionTransform.iceberg_bucket(n)
        return cls._from_py_partition_transform(pt)

    @classmethod
    def iceberg_truncate(cls, w: int) -> PartitionTransform:
        """Creates an Iceberg truncate partition transform that truncates values to width w.

        Args:
            w: Width to truncate values to.

        Returns:
            PartitionTransform: A new Iceberg truncate partition transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_truncate(10)
            >>> assert transform.is_iceberg_truncate()
            >>> assert transform.width == 10
        """
        pt = PyPartitionTransform.iceberg_truncate(w)
        return cls._from_py_partition_transform(pt)

    def is_identity(self) -> bool:
        """Checks if this is an identity partition transform.

        Returns:
            bool: True if this is an identity transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.identity()
            >>> assert transform.is_identity()
        """
        return self._partition_transform.is_identity()

    def is_year(self) -> bool:
        """Checks if this is a year partition transform.

        Returns:
            bool: True if this is a year transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.year()
            >>> assert transform.is_year()
        """
        return self._partition_transform.is_year()

    def is_month(self) -> bool:
        """Checks if this is a month partition transform.

        Returns:
            bool: True if this is a month transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.month()
            >>> assert transform.is_month()
        """
        return self._partition_transform.is_month()

    def is_day(self) -> bool:
        """Checks if this is a day partition transform.

        Returns:
            bool: True if this is a day transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.day()
            >>> assert transform.is_day()
        """
        return self._partition_transform.is_day()

    def is_hour(self) -> bool:
        """Checks if this is an hour partition transform.

        Returns:
            bool: True if this is an hour transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.hour()
            >>> assert transform.is_hour()
        """
        return self._partition_transform.is_hour()

    def is_iceberg_bucket(self) -> bool:
        """Checks if this is an Iceberg bucket partition transform.

        Returns:
            bool: True if this is an Iceberg bucket transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_bucket(10)
            >>> assert transform.is_iceberg_bucket()
        """
        return self._partition_transform.is_iceberg_bucket()

    def is_iceberg_truncate(self) -> bool:
        """Checks if this is an Iceberg truncate partition transform.

        Returns:
            bool: True if this is an Iceberg truncate transform.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_truncate(10)
            >>> assert transform.is_iceberg_truncate()
        """
        return self._partition_transform.is_iceberg_truncate()

    @property
    def num_buckets(self) -> int:
        """If this is an iceberg_bucket partition transform, then returns the number of buckets; otherwise an attribute error is raised.

        Returns:
            int: Number of buckets.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_bucket(10)
            >>> assert transform.num_buckets == 10
        """
        return self._partition_transform.num_buckets()

    @property
    def width(self) -> int:
        """If this is an iceberg_truncate partition transform, then returns the truncation width; otherwise an attribute error is raised.

        Returns:
            int: Truncation width.

        Examples:
            >>> from daft.io.partitioning import PartitionTransform
            >>> transform = PartitionTransform.iceberg_truncate(10)
            >>> assert transform.width == 10
        """
        return self._partition_transform.width()
